
import warnings
from typing import List, Optional, Union

import cadquery as cq
import paramak


class SubmersionTokamak(paramak.Reactor):
    """Creates geometry for a simple submersion reactor including a plasma,
    cylindrical center column shielding, inboard and outboard breeder blanket,
    divertor (upper and lower), support legs. Optional coat hanger shaped
    toroidal field coils and pf coils.

    Arguments:
        inner_bore_radial_thickness: the radial thickness of the inner bore
            (cm)
        inboard_tf_leg_radial_thickness: the radial thickness of the inner leg
            of the toroidal field coils (cm)
        center_column_shield_radial_thickness: the radial thickness of the
            center column shield (cm)
        inboard_blanket_radial_thickness: the radial thickness of the inboard
            blanket (cm)
        firstwall_radial_thickness: the radial thickness of the first wall (cm)
        inner_plasma_gap_radial_thickness: the radial thickness of the inboard
            gap between the plasma and the center column shield (cm)
        plasma_radial_thickness: the radial thickness of the plasma (cm)
        divertor_radial_thickness: the radial thickness of the divertors (cm)
        support_radial_thickness: the radial thickness of the upper and lower
            supports (cm)
        outer_plasma_gap_radial_thickness: the radial thickness of the outboard
            gap between the plasma and the first wall (cm)
        outboard_blanket_radial_thickness: the radial thickness of the blanket
            (cm)
        blanket_rear_wall_radial_thickness: the radial thickness of the rear
            wall of the blanket (cm)
        elongation: the elongation of the plasma
        triangularity: the triangularity of the plasma
        number_of_tf_coils: the number of tf coils.
        rotation_angle: the angle of the sector that is desired.
        outboard_tf_coil_radial_thickness: the radial thickness of the toroidal
            field coil.
        rear_blanket_to_tf_gap: the radial distance between the rear of the
            blanket and the toroidal field coil.
        outboard_tf_coil_poloidal_thickness: the vertical thickness of each
            poloidal field coil.
        pf_coil_vertical_thicknesses: the vertical thickness of each poloidal
            field coil.
        pf_coil_radial_thicknesses: the radial thickness of  each poloidal
            field coil.
        divertor_position: the position of the divertor, "upper", "lower" or
            "both". Defaults to "both".
        support_position: the position of the supports, "upper", "lower" or
            "both". Defaults to "both".
    """

    def __init__(
        self,
        inner_bore_radial_thickness: float,
        inboard_tf_leg_radial_thickness: float,
        center_column_shield_radial_thickness: float,
        inboard_blanket_radial_thickness: float,
        firstwall_radial_thickness: float,
        inner_plasma_gap_radial_thickness: float,
        plasma_radial_thickness: float,
        divertor_radial_thickness: float,
        support_radial_thickness: float,
        outer_plasma_gap_radial_thickness: float,
        outboard_blanket_radial_thickness: float,
        blanket_rear_wall_radial_thickness: float,
        elongation: float,
        triangularity: float,
        number_of_tf_coils: int = 16,
        rotation_angle: float = 360.0,
        outboard_tf_coil_radial_thickness: Optional[float] = None,
        rear_blanket_to_tf_gap: Optional[float] = None,
        outboard_tf_coil_poloidal_thickness: Optional[float] = None,
        pf_coil_vertical_thicknesses: Optional[float] = None,
        pf_coil_radial_thicknesses: Optional[float] = None,
        pf_coil_radial_position: Optional[Union[float, List[float]]] = None,
        pf_coil_vertical_position: Optional[Union[float, List[float]]] = None,
        pf_coil_case_thicknesses: Optional[float] = 10,
        divertor_position: Optional[str] = "both",
        support_position: Optional[str] = "both",
    ):

        super().__init__([])

        self.inner_bore_radial_thickness = inner_bore_radial_thickness
        self.inboard_tf_leg_radial_thickness = inboard_tf_leg_radial_thickness
        self.center_column_shield_radial_thickness = (
            center_column_shield_radial_thickness
        )
        self.inboard_blanket_radial_thickness = \
            inboard_blanket_radial_thickness
        self.firstwall_radial_thickness = firstwall_radial_thickness
        self.inner_plasma_gap_radial_thickness = \
            inner_plasma_gap_radial_thickness
        self.plasma_radial_thickness = plasma_radial_thickness
        self.outer_plasma_gap_radial_thickness = \
            outer_plasma_gap_radial_thickness
        self.outboard_blanket_radial_thickness = \
            outboard_blanket_radial_thickness
        self.blanket_rear_wall_radial_thickness = \
            blanket_rear_wall_radial_thickness
        self.pf_coil_radial_thicknesses = pf_coil_radial_thicknesses
        self.outboard_tf_coil_radial_thickness = \
            outboard_tf_coil_radial_thickness
        self.outboard_tf_coil_poloidal_thickness = \
            outboard_tf_coil_poloidal_thickness
        self.divertor_radial_thickness = divertor_radial_thickness
        self.support_radial_thickness = support_radial_thickness
        self.elongation = elongation
        self.triangularity = triangularity
        self.rear_blanket_to_tf_gap = \
            rear_blanket_to_tf_gap
        self.pf_coil_vertical_thicknesses = pf_coil_vertical_thicknesses
        self.pf_coil_case_thicknesses = pf_coil_case_thicknesses
        self.number_of_tf_coils = number_of_tf_coils
        self.rotation_angle = rotation_angle
        self.divertor_position = divertor_position
        self.support_position = support_position
        self.pf_coil_vertical_position = pf_coil_vertical_position
        self.pf_coil_radial_position = pf_coil_radial_position

        # adds self.input_variable_names from the Reactor class
        self.input_variable_names = self.input_variable_names + [
            'inner_bore_radial_thickness',
            'inboard_tf_leg_radial_thickness',
            'center_column_shield_radial_thickness',
            'inboard_blanket_radial_thickness',
            'firstwall_radial_thickness',
            'inner_plasma_gap_radial_thickness',
            'plasma_radial_thickness',
            'divertor_radial_thickness',
            'support_radial_thickness',
            'outer_plasma_gap_radial_thickness',
            'outboard_blanket_radial_thickness',
            'blanket_rear_wall_radial_thickness',
            'elongation',
            'triangularity',
            'number_of_tf_coils',
            'rotation_angle',
            'outboard_tf_coil_radial_thickness',
            'rear_blanket_to_tf_gap',
            'outboard_tf_coil_poloidal_thickness',
            'pf_coil_vertical_thicknesses',
            'pf_coil_radial_thicknesses',
            'pf_coil_radial_position',
            'pf_coil_vertical_position',
            'pf_coil_case_thicknesses',
            'divertor_position',
            'support_position',
        ]

        # set by make_plasma
        self.major_radius = None
        self.minor_radius = None

    @property
    def pf_coil_radial_position(self):
        return self._pf_coil_radial_position

    @pf_coil_radial_position.setter
    def pf_coil_radial_position(self, value):
        if not isinstance(value, list) and value is not None:
            raise ValueError("pf_coil_radial_position must be a list")
        self._pf_coil_radial_position = value

    @property
    def pf_coil_radial_thicknesses(self):
        return self._pf_coil_radial_thicknesses

    @pf_coil_radial_thicknesses.setter
    def pf_coil_radial_thicknesses(self, value):
        if not isinstance(value, list) and value is not None:
            raise ValueError("pf_coil_radial_thicknesses must be a list")
        self._pf_coil_radial_thicknesses = value

    @property
    def pf_coil_vertical_thicknesses(self):
        return self._pf_coil_vertical_thicknesses

    @pf_coil_vertical_thicknesses.setter
    def pf_coil_vertical_thicknesses(self, value):
        if not isinstance(value, list) and value is not None:
            raise ValueError("pf_coil_vertical_thicknesses must be a list")
        self._pf_coil_vertical_thicknesses = value

    @property
    def divertor_position(self):
        return self._divertor_position

    @divertor_position.setter
    def divertor_position(self, value):
        acceptable_values = ["upper", "lower", "both"]
        if value in acceptable_values:
            self._divertor_position = value
        else:
            msg = "divertor_position must be 'upper', 'lower' or 'both'"
            raise ValueError(msg)

    @property
    def support_position(self):
        return self._support_position

    @support_position.setter
    def support_position(self, value):
        acceptable_values = ["upper", "lower", "both"]
        if value in acceptable_values:
            self._support_position = value
        else:
            msg = "support_position must be 'upper', 'lower' or 'both'"
            raise ValueError(msg)

    def create_solids(self):
        """Creates a list of paramak.Shape for components and saves it in
        self.shapes_and_components
        """

        uncut_shapes = []

        self._rotation_angle_check()
        uncut_shapes.append(self._make_plasma())
        self._make_radial_build()
        self._make_vertical_build()
        uncut_shapes.append(self._make_center_column_shield())
        uncut_shapes.append(self._make_firstwall())
        uncut_shapes.append(self._make_blanket())
        uncut_shapes.append(self._make_divertor())
        uncut_shapes.append(self._make_supports())
        uncut_shapes.append(self._make_rear_blanket_wall())
        uncut_shapes += self._make_tf_coils()
        pf_coils = self._make_pf_coils()

        if pf_coils is None:
            shapes_and_components = uncut_shapes
        else:
            for shape in uncut_shapes:
                for pf_coil in pf_coils:
                    shape.solid = shape.solid.cut(pf_coil.solid)
            shapes_and_components = pf_coils + uncut_shapes

        self.shapes_and_components = shapes_and_components

    def _rotation_angle_check(self):

        if self.rotation_angle == 360:
            msg = "360 degree rotation may result" + \
                " in a Standard_ConstructionError or AttributeError"
            warnings.warn(msg, UserWarning)

    def _make_radial_build(self):

        # this is the radial build sequence, where one component stops and
        # another starts

        self._inner_bore_start_radius = 0
        self._inner_bore_end_radius = (
            self._inner_bore_start_radius + self.inner_bore_radial_thickness
        )

        self._inboard_tf_coils_start_radius = self._inner_bore_end_radius
        self._inboard_tf_coils_end_radius = (
            self._inboard_tf_coils_start_radius +
            self.inboard_tf_leg_radial_thickness)

        self._center_column_shield_start_radius = \
            self._inboard_tf_coils_end_radius
        self._center_column_shield_end_radius = (
            self._center_column_shield_start_radius
            + self.center_column_shield_radial_thickness
        )

        self._inboard_blanket_start_radius = \
            self._center_column_shield_end_radius
        self._inboard_blanket_end_radius = (
            self._inboard_blanket_start_radius +
            self.inboard_blanket_radial_thickness)

        self._inboard_firstwall_start_radius = self._inboard_blanket_end_radius
        self._inboard_firstwall_end_radius = (
            self._inboard_firstwall_start_radius +
            self.firstwall_radial_thickness)

        self._inner_plasma_gap_start_radius = \
            self._inboard_firstwall_end_radius
        self._inner_plasma_gap_end_radius = (
            self._inner_plasma_gap_start_radius +
            self.inner_plasma_gap_radial_thickness)

        self._plasma_start_radius = self._inner_plasma_gap_end_radius
        self._plasma_end_radius = \
            self._plasma_start_radius + \
            self.plasma_radial_thickness

        self._outer_plasma_gap_start_radius = self._plasma_end_radius
        self._outer_plasma_gap_end_radius = (
            self._outer_plasma_gap_start_radius +
            self.outer_plasma_gap_radial_thickness)

        self._outboard_firstwall_start_radius = \
            self._outer_plasma_gap_end_radius
        self._outboard_firstwall_end_radius = (
            self._outboard_firstwall_start_radius +
            self.firstwall_radial_thickness)

        self._outboard_blanket_start_radius = \
            self._outboard_firstwall_end_radius
        self._outboard_blanket_end_radius = (
            self._outboard_blanket_start_radius +
            self.outboard_blanket_radial_thickness)

        self._blanket_rear_wall_start_radius = \
            self._outboard_blanket_end_radius
        self._blanket_rear_wall_end_radius = (
            self._blanket_rear_wall_start_radius +
            self.blanket_rear_wall_radial_thickness)

        self._tf_info_provided = False
        if (
            self.outboard_tf_coil_radial_thickness is not None
            and self.rear_blanket_to_tf_gap is not None
            and self.outboard_tf_coil_poloidal_thickness is not None
        ):
            self._tf_info_provided = True
            self._outboard_tf_coil_start_radius = (
                self._blanket_rear_wall_end_radius +
                self.rear_blanket_to_tf_gap)
            self._outboard_tf_coil_end_radius = (
                self._outboard_tf_coil_start_radius +
                self.outboard_tf_coil_radial_thickness)

        self._divertor_start_radius = (
            self._plasma.high_point[0] - 0.5 * self.divertor_radial_thickness
        )
        self._divertor_end_radius = (
            self._plasma.high_point[0] + 0.5 * self.divertor_radial_thickness
        )

        self._support_start_radius = (
            self._plasma.high_point[0] - 0.5 * self.support_radial_thickness
        )
        self._support_end_radius = (
            self._plasma.high_point[0] + 0.5 * self.support_radial_thickness
        )

    def _make_vertical_build(self):

        # this is the vertical build sequence, componets build on each other in
        # a similar manner to the radial build

        self._plasma_start_height = 0
        self._plasma_end_height = self._plasma.high_point[1]

        self._plasma_to_divertor_gap_start_height = self._plasma_end_height
        self._plasma_to_divertor_gap_end_height = (
            self._plasma_to_divertor_gap_start_height +
            self.outer_plasma_gap_radial_thickness)

        # the firstwall is cut by the divertor but uses the same control points
        self._firstwall_start_height = self._plasma_to_divertor_gap_end_height
        self._firstwall_end_height = self._firstwall_start_height + \
            self.firstwall_radial_thickness

        self._blanket_start_height = self._firstwall_end_height
        self._blanket_end_height = (
            self._blanket_start_height + self.outboard_blanket_radial_thickness
        )

        self._blanket_rear_wall_start_height = self._blanket_end_height
        self._blanket_rear_wall_end_height = (
            self._blanket_rear_wall_start_height +
            self.blanket_rear_wall_radial_thickness)

        if self._tf_info_provided:
            self._outboard_tf_coils_vertical_height = \
                self._blanket_rear_wall_end_height * 1.5
            self._outboard_tf_coils_horizontal_length = \
                self._blanket_rear_wall_end_radius * 0.75

    def _make_center_column_shield(self):

        self._center_column_shield = paramak.CenterColumnShieldCylinder(
            height=self._blanket_rear_wall_end_height * 2,
            inner_radius=self._center_column_shield_start_radius,
            outer_radius=self._center_column_shield_end_radius,
            rotation_angle=self.rotation_angle,
            name="center_column_shield",
        )
        return self._center_column_shield

    def _make_plasma(self):

        # sets major radius and minor radius from equatorial_points to allow a
        # radial build this helps avoid the plasma overlapping the center
        # column and other components
        inner_equatorial_point = (
            self.inner_bore_radial_thickness
            + self.inboard_tf_leg_radial_thickness
            + self.center_column_shield_radial_thickness
            + self.inboard_blanket_radial_thickness
            + self.firstwall_radial_thickness
            + self.inner_plasma_gap_radial_thickness
        )
        outer_equatorial_point = \
            inner_equatorial_point + self.plasma_radial_thickness
        self.major_radius = \
            (outer_equatorial_point + inner_equatorial_point) / 2
        self.minor_radius = self.major_radius - inner_equatorial_point

        plasma = paramak.Plasma(
            major_radius=self.major_radius,
            minor_radius=self.minor_radius,
            elongation=self.elongation,
            triangularity=self.triangularity,
            rotation_angle=self.rotation_angle,
        )

        self._plasma = plasma
        return self._plasma

    def _make_firstwall(self):

        # this is used to cut the inboard blanket and then fused / unioned with
        # the firstwall
        self._inboard_firstwall = paramak.BlanketFP(
            plasma=self._plasma,
            offset_from_plasma=self.inner_plasma_gap_radial_thickness,
            start_angle=90,
            stop_angle=270,
            thickness=self.firstwall_radial_thickness,
            rotation_angle=self.rotation_angle,
            color=(0.5, 0.5, 0.5),
        )

        self._firstwall = paramak.BlanketFP(
            plasma=self._plasma,
            offset_from_plasma=self.outer_plasma_gap_radial_thickness,
            start_angle=90,
            stop_angle=-90,
            thickness=self.firstwall_radial_thickness,
            rotation_angle=self.rotation_angle,
            name="outboard_firstwall",
            union=self._inboard_firstwall,
            color=(0.5, 0.5, 0.5),
        )
        return self._firstwall

    def _make_divertor(self):
        fw_envelope_inboard = paramak.BlanketFP(
            plasma=self._plasma,
            offset_from_plasma=self.inner_plasma_gap_radial_thickness,
            start_angle=90,
            stop_angle=270,
            thickness=self.firstwall_radial_thickness,
            rotation_angle=self.rotation_angle,
        )

        fw_envelope = paramak.BlanketFP(
            plasma=self._plasma,
            offset_from_plasma=self.outer_plasma_gap_radial_thickness,
            start_angle=90,
            stop_angle=-90,
            thickness=self.firstwall_radial_thickness,
            rotation_angle=self.rotation_angle,
            name="outboard_firstwall",
            union=fw_envelope_inboard,
        )
        divertor_height = self._blanket_rear_wall_end_height

        divertor_height_top = divertor_height
        divertor_height_bottom = -divertor_height

        if self.divertor_position == "lower":
            divertor_height_top = 0
        elif self.divertor_position == "upper":
            divertor_height_bottom = 0

        self._divertor = paramak.RotateStraightShape(
            points=[
                (self._divertor_start_radius, divertor_height_bottom),
                (self._divertor_end_radius, divertor_height_bottom),
                (self._divertor_end_radius, divertor_height_top),
                (self._divertor_start_radius, divertor_height_top)
            ],
            intersect=fw_envelope,
            rotation_angle=self.rotation_angle,
            name="divertor",
            color=(1., 0.667, 0.),
        )

        self._firstwall.cut = self._divertor
        self._inboard_firstwall.cut = self._divertor
        return self._divertor

    def _make_blanket(self):
        self._inboard_blanket = paramak.CenterColumnShieldCylinder(
            height=self._blanket_end_height * 2,
            inner_radius=self._inboard_blanket_start_radius,
            outer_radius=max(self._inboard_firstwall.points)[0],
            rotation_angle=self.rotation_angle,
            cut=self._inboard_firstwall,
        )

        # this takes a single solid from a compound of solids by finding the
        # solid nearest to a point
        # TODO: find alternative
        self._inboard_blanket.solid = self._inboard_blanket.solid.solids(
            cq.selectors.NearestToPointSelector((0, 0, 0))
        )

        # this is the outboard fused /unioned with the inboard blanket

        self._blanket = paramak.BlanketFP(
            plasma=self._plasma,
            start_angle=90,
            stop_angle=-90,
            offset_from_plasma=self.outer_plasma_gap_radial_thickness
            + self.firstwall_radial_thickness,
            thickness=self.outboard_blanket_radial_thickness,
            rotation_angle=self.rotation_angle,
            name="blanket",
            color=(0., 1., 0.498),
            union=self._inboard_blanket,
        )
        return self._blanket

    def _make_supports(self):
        blanket_envelope = paramak.BlanketFP(
            plasma=self._plasma,
            start_angle=90,
            stop_angle=-90,
            offset_from_plasma=self.outer_plasma_gap_radial_thickness
            + self.firstwall_radial_thickness,
            thickness=self.outboard_blanket_radial_thickness,
            rotation_angle=self.rotation_angle,
            union=self._inboard_blanket,
        )
        support_height = self._blanket_rear_wall_end_height
        support_height_top = support_height
        support_height_bottom = -support_height

        if self.support_position == "lower":
            support_height_top = 0
        elif self.support_position == "upper":
            support_height_bottom = 0

        self._supports = paramak.RotateStraightShape(
            points=[
                (self._support_start_radius, support_height_bottom),
                (self._support_end_radius, support_height_bottom),
                (self._support_end_radius, support_height_top),
                (self._support_start_radius, support_height_top)
            ],
            rotation_angle=self.rotation_angle,
            name="supports",
            color=(0., 0., 0.),
            intersect=blanket_envelope,
        )

        self._blanket.solid = self._blanket.solid.cut(self._supports.solid)

        return self._supports

    def _make_rear_blanket_wall(self):
        self._outboard_rear_blanket_wall_upper = paramak.RotateStraightShape(
            points=[
                (
                    self._center_column_shield_end_radius,
                    self._blanket_rear_wall_start_height
                ),
                (
                    self._center_column_shield_end_radius,
                    self._blanket_rear_wall_end_height
                ),
                (
                    max(self._inboard_firstwall.points)[0],
                    self._blanket_rear_wall_end_height,
                ),
                (
                    max(self._inboard_firstwall.points)[0],
                    self._blanket_rear_wall_start_height,
                ),
            ],
            rotation_angle=self.rotation_angle,
        )

        self._outboard_rear_blanket_wall_lower = paramak.RotateStraightShape(
            points=[
                (
                    self._center_column_shield_end_radius,
                    -self._blanket_rear_wall_start_height
                ),
                (
                    self._center_column_shield_end_radius,
                    -self._blanket_rear_wall_end_height
                ),
                (
                    max(self._inboard_firstwall.points)[0],
                    -self._blanket_rear_wall_end_height,
                ),
                (
                    max(self._inboard_firstwall.points)[0],
                    -self._blanket_rear_wall_start_height,
                ),
            ],
            rotation_angle=self.rotation_angle,
        )

        self._outboard_rear_blanket_wall = paramak.BlanketFP(
            plasma=self._plasma,
            start_angle=90,
            stop_angle=-90,
            offset_from_plasma=self.outer_plasma_gap_radial_thickness +
            self.firstwall_radial_thickness +
            self.outboard_blanket_radial_thickness,
            thickness=self.blanket_rear_wall_radial_thickness,
            rotation_angle=self.rotation_angle,
            name="outboard_rear_blanket_wall",
            color=(0., 1., 1.),
            union=[
                self._outboard_rear_blanket_wall_upper,
                self._outboard_rear_blanket_wall_lower],
        )

        return self._outboard_rear_blanket_wall

    def _make_pf_coils(self):
        if None not in [self.pf_coil_vertical_thicknesses,
                        self.pf_coil_radial_thicknesses,
                        self.pf_coil_vertical_position,
                        self.pf_coil_radial_position]:
            list_of_components = []

            # TODO make use of counter in the name attribute

            center_points = [
                (x, y) for x, y in zip(
                    self.pf_coil_radial_position, self.pf_coil_vertical_position)]

            self._pf_coils = self._pf_coil = paramak.PoloidalFieldCoilSet(
                heights=self.pf_coil_vertical_thicknesses,
                widths=self.pf_coil_radial_thicknesses,
                center_points=center_points,
                rotation_angle=self.rotation_angle,
                name="pf_coil",
            )
            list_of_components.append(self._pf_coil)

            if self.pf_coil_case_thicknesses is not None:
                self._pf_coils_casing = paramak.PoloidalFieldCoilCaseSetFC(
                    pf_coils=self._pf_coil,
                    casing_thicknesses=self.pf_coil_case_thicknesses,
                    rotation_angle=self.rotation_angle,
                    name="pf_coil_case",
                )
                list_of_components.append(self._pf_coils_casing)

            return list_of_components
        else:
            print(
                'pf_coil_vertical_thicknesses, pf_coil_radial_thicknesses, '
                'pf_coil_radial_position, pf_coil_vertical_position not '
                'so not making pf coils')
            return None

    def _make_tf_coils(self):
        list_of_components = []

        self._inboard_tf_coils = paramak.CenterColumnShieldCylinder(
            height=self._blanket_rear_wall_end_height * 2,
            inner_radius=self._inboard_tf_coils_start_radius,
            outer_radius=self._inboard_tf_coils_end_radius,
            rotation_angle=self.rotation_angle,
            name="inboard_tf_coils",
            color=(0, 0, 1),
        )
        list_of_components.append(self._inboard_tf_coils)

        if None not in [
                self.rear_blanket_to_tf_gap,
                self.outboard_tf_coil_radial_thickness,
                self.outboard_tf_coil_poloidal_thickness,
                self.number_of_tf_coils] and self.number_of_tf_coils > 1:

            if self._tf_info_provided:
                self._tf_coil = paramak.ToroidalFieldCoilCoatHanger(
                    with_inner_leg=False,
                    horizontal_start_point=(
                        self._inboard_tf_coils_start_radius,
                        self._blanket_rear_wall_end_height,
                    ),
                    vertical_mid_point=(self._outboard_tf_coil_start_radius, 0),
                    thickness=self.outboard_tf_coil_radial_thickness,
                    number_of_coils=self.number_of_tf_coils,
                    distance=self.outboard_tf_coil_poloidal_thickness,
                    rotation_angle=self.rotation_angle,
                    horizontal_length=self._outboard_tf_coils_horizontal_length,
                    vertical_length=self._outboard_tf_coils_vertical_height
                )
                list_of_components.append(self._tf_coil)

        return list_of_components




import unittest

import paramak


class TestCenterColumnShieldFlatTopCircular(unittest.TestCase):

    def setUp(self):
        self.test_shape = paramak.CenterColumnShieldFlatTopCircular(
            height=600, arc_height=400, inner_radius=100, mid_radius=150, outer_radius=200)

    def test_default_parameters(self):
        """Checks that the default parameters of a CenterColumnShieldFlatTopCircular are
        correct."""

        assert self.test_shape.rotation_angle == 360

    def test_points_calculation(self):
        """Checks that the points used to construct the CenterColumnShieldFlatTopCircular
        component are calculated correctly from the parameters given."""

        assert self.test_shape.points == [
            (100, 0, "straight"),
            (100, 300, "straight"),
            (200, 300, "straight"),
            (200, 200, "circle"),
            (150, 0, "circle"),
            (200, -200, "straight"),
            (200, -300, "straight"),
            (100, -300, "straight"),
        ]

    def test_processed_points_calculation(self):
        """Checks that the processed_points used to construct the
        CenterColumnShieldFlatTopCircular component are calculated correctly
        from the parameters given."""

        assert self.test_shape.processed_points == [
            (100, 0, "straight"),
            (100, 300, "straight"),
            (200, 300, "straight"),
            (200, 200, "circle"),
            (150, 0, "circle"),
            (200, -200, "straight"),
            (200, -300, "straight"),
            (100, -300, "straight"),
            (100, 0, "straight")
        ]

    def test_creation(self):
        """Creates a center column shield using the
        CenterColumnShieldFlatTopCircular parametric component and checks that
        a cadquery solid is created."""

        assert self.test_shape.solid is not None
        assert self.test_shape.volume() > 1000

    def test_faces(self):
        """Creates a center column shield using the
        CenterColumnShieldFlatTopCircular parametric component and checks that
        a solid is created with the correct number of faces"""

        assert len(self.test_shape.areas) == 6
        assert len(set([round(i) for i in self.test_shape.areas])) == 4

        self.test_shape.rotation_angle = 180
        assert len(self.test_shape.areas) == 8
        assert len(set([round(i) for i in self.test_shape.areas])) == 5




import math

from paramak import RotateStraightShape
from paramak.utils import rotate


class RotatedIsoscelesTriangle(RotateStraightShape):
    """Creates a rotated triangle (truncated triangle) shape.

    Args:
        base_length (float): the length of the base of the triangle (cm).
        height (float): the height of the triangle (cm).
        pivot_point ((float, float)): the coordinates of the tip of the
            triangle at the opposite side to the base of the triangle.
        pivot_angle (float, optional): the angle (in degrees) to pivot (rotate)
            the shape by around the pivot point. Defaults to 0.
        name (str, optional): defaults to "rotated_triangle".
    """

    def __init__(
        self,
        base_length,
        height,
        pivot_point,
        pivot_angle=0.,
        name="rotated_triangle",
        **kwargs
    ):

        super().__init__(
            name=name,
            **kwargs
        )

        self.base_length = base_length
        self.height = height
        self.pivot_point = pivot_point
        self.pivot_angle = pivot_angle

    @property
    def pivot_point(self):
        return self._pivot_point

    @pivot_point.setter
    def pivot_point(self, pivot_point):
        self._pivot_point = pivot_point

    @property
    def base_length(self):
        return self._base_length

    @base_length.setter
    def base_length(self, base_length):
        self._base_length = base_length

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, height):
        self._height = height

    @property
    def pivot_angle(self):
        return self._pivot_angle

    @pivot_angle.setter
    def pivot_angle(self, pivot_angle):
        self._pivot_angle = pivot_angle

    def find_points(self):
        """Finds the XZ points joined by straight connections that describe
        the 2D profile of the triangle shape."""

        non_rotated_points = [
            self.pivot_point,
            (
                self.pivot_point[0] - self.base_length / 2.0,
                self.pivot_point[1] - self.height
            ),
            (
                self.pivot_point[0] + self.base_length / 2.0,
                self.pivot_point[1] - self.height
            ),
        ]

        points = []

        for point in non_rotated_points:
            x, y = rotate(
                origin=self.pivot_point,
                point=point,
                angle=math.radians(self.pivot_angle)
            )
            points.append((x, y))

        self.points = points




from typing import Optional, Tuple

import numpy as np
from paramak import ExtrudeMixedShape
from paramak.utils import add_thickness
from scipy import integrate
from scipy.optimize import minimize


class ToroidalFieldCoilPrincetonD(ExtrudeMixedShape):
    """Toroidal field coil based on Princeton-D curve

    Args:
        R1: smallest radius (cm)
        R2: largest radius (cm)
        thickness: magnet thickness (cm)
        distance: extrusion distance (cm)
        number_of_coils: the number of tf coils. This changes by the
            azimuth_placement_angle dividing up 360 degrees by the number of
            coils.
        vertical_displacement: vertical displacement (cm). Defaults to 0.0.
        with_inner_leg: Include the inner tf leg. Defaults to True.
    """

    def __init__(
        self,
        R1: float,
        R2: float,
        thickness: float,
        distance: float,
        number_of_coils: int,
        vertical_displacement: Optional[float] = 0.0,
        with_inner_leg: Optional[bool] = True,
        color: Optional[Tuple[float, float, float, Optional[float]]] = (0., 0., 1.),
        **kwargs
    ) -> None:

        super().__init__(
            distance=distance,
            color=color,
            **kwargs
        )

        self.R1 = R1
        self.R2 = R2
        self.thickness = thickness
        self.distance = distance
        self.number_of_coils = number_of_coils
        self.vertical_displacement = vertical_displacement
        self.with_inner_leg = with_inner_leg

    @property
    def inner_points(self):
        self.points
        return self._inner_points

    @inner_points.setter
    def inner_points(self, value):
        self._inner_points = value

    @property
    def outer_points(self):
        self.points
        return self._outer_points

    @outer_points.setter
    def outer_points(self, value):
        self._outer_points = value

    @property
    def azimuth_placement_angle(self):
        self.find_azimuth_placement_angle()
        return self._azimuth_placement_angle

    @azimuth_placement_angle.setter
    def azimuth_placement_angle(self, value):
        self._azimuth_placement_angle = value

    def _compute_inner_points(self, R1, R2):
        """Computes the inner curve points

        Args:
            R1 (float): smallest radius (cm)
            R2 (float): largest radius (cm)

        Returns:
            (list, list, list): R, Z and derivative lists for outer curve
            points
        """
        def error(z_0, R0, R2):
            segment = get_segment(R0, R2, z_0)
            return abs(segment[1][-1])

        def get_segment(a, b, z_0):
            a_R = np.linspace(a, b, num=70, endpoint=True)
            asol = integrate.odeint(solvr, [z_0, 0], a_R)
            return a_R, asol[:, 0], asol[:, 1]

        def solvr(Y, R):
            return [Y[1], -1 / (k * R) * (1 + Y[1]**2)**(3 / 2)]

        R0 = (R1 * R2)**0.5
        k = 0.5 * np.log(R2 / R1)

        # computing of z_0
        # z_0 is computed by ensuring outer segment end is zero
        z_0 = 10  # initial guess for z_0
        res = minimize(error, z_0, args=(R0, R2))
        z_0 = res.x

        # compute inner and outer segments
        segment1 = get_segment(R0, R1, z_0)
        segment2 = get_segment(R0, R2, z_0)

        R = np.concatenate([np.flip(segment1[0]), segment2[0]
                            [1:], np.flip(segment2[0])[1:], segment1[0][1:]])
        Z = np.concatenate([np.flip(segment1[1]), segment2[1]
                            [1:], -np.flip(segment2[1])[1:], -segment1[1][1:]])
        return R, Z

    def find_points(self):
        """Finds the XZ points joined by connections that describe the 2D
        profile of the toroidal field coil shape."""
        # compute inner points
        R_inner, Z_inner = self._compute_inner_points(
            self.R1 + self.thickness, self.R2)

        # compute outer points
        dz_dr = np.diff(Z_inner) / np.diff(R_inner)
        dz_dr[0] = float("-inf")
        dz_dr = np.append(dz_dr, float("inf"))
        R_outer, Z_outer = add_thickness(
            R_inner, Z_inner, self.thickness, dy_dx=dz_dr
        )
        R_outer, Z_outer = np.flip(R_outer), np.flip(Z_outer)

        # add vertical displacement
        Z_outer += self.vertical_displacement
        Z_inner += self.vertical_displacement

        # extract helping points for inner leg
        inner_leg_connection_points = [
            (R_inner[0], Z_inner[0]),
            (R_inner[-1], Z_inner[-1]),
            (R_outer[0], Z_outer[0]),
            (R_outer[-1], Z_outer[-1])
        ]
        self.inner_leg_connection_points = inner_leg_connection_points

        # add the leg to the points
        if self.with_inner_leg:
            R_inner = np.append(R_inner, R_inner[0])
            Z_inner = np.append(Z_inner, Z_inner[0])

            R_outer = np.append(R_outer, R_outer[0])
            Z_outer = np.append(Z_outer, Z_outer[0])
        # add connections
        inner_points = [[r, z, 'spline'] for r, z in zip(R_inner, Z_inner)]
        outer_points = [[r, z, 'spline'] for r, z in zip(R_outer, Z_outer)]
        if self.with_inner_leg:
            outer_points[-2][2] = 'straight'
            inner_points[-2][2] = 'straight'

        inner_points[-1][2] = 'straight'
        outer_points[-1][2] = 'straight'

        points = inner_points + outer_points
        self.outer_points = np.vstack((R_outer, Z_outer)).T
        self.inner_points = np.vstack((R_inner, Z_inner)).T
        self.points = points

    def find_azimuth_placement_angle(self):
        """Calculates the azimuth placement angles based on the number of tf
        coils"""

        angles = list(
            np.linspace(
                0,
                360,
                self.number_of_coils,
                endpoint=False))

        self.azimuth_placement_angle = angles




import os
import unittest
import pytest
from pathlib import Path

import paramak


class TestPlasma(unittest.TestCase):
    def test_plasma_attributes(self):
        """Creates a plasma object using the Plasma parametric component and
        checks that its attributes can be set correctly."""

        test_plasma = paramak.Plasma()

        assert isinstance(test_plasma.elongation, float)

        def test_plasma_elongation_min_setting():
            """checks ValueError is raised when an elongation < 0 is specified"""

            test_plasma.elongation = -1

        self.assertRaises(ValueError, test_plasma_elongation_min_setting)

        def test_plasma_elongation_max_setting():
            """checks ValueError is raised when an elongation > 4 is specified"""

            test_plasma.elongation = 400

        self.assertRaises(ValueError, test_plasma_elongation_max_setting)

        def minor_radius_out_of_range():
            """checks ValueError is raised when an minor_radius < 1 is
            specified"""

            test_plasma.minor_radius = 0.5

        self.assertRaises(ValueError, minor_radius_out_of_range)

        def major_radius_out_of_range():
            """checks ValueError is raised when an manor_radius < 1 is
            specified"""

            test_plasma.major_radius = 0.5

        self.assertRaises(ValueError, major_radius_out_of_range)

    def test_plasma_points_of_interest(self):
        test_plasma = paramak.Plasma(vertical_displacement=2)
        assert test_plasma.high_point == (
            test_plasma.major_radius -
            test_plasma.triangularity * test_plasma.minor_radius,
            test_plasma.elongation * test_plasma.minor_radius +
            test_plasma.vertical_displacement,
        )
        assert test_plasma.low_point == (
            test_plasma.major_radius -
            test_plasma.triangularity * test_plasma.minor_radius,
            -test_plasma.elongation * test_plasma.minor_radius +
            test_plasma.vertical_displacement,
        )
        assert test_plasma.outer_equatorial_point == (
            test_plasma.major_radius + test_plasma.minor_radius,
            test_plasma.vertical_displacement
        )
        assert test_plasma.inner_equatorial_point == (
            test_plasma.major_radius - test_plasma.minor_radius,
            test_plasma.vertical_displacement
        )

    def test_plasma_x_points(self):
        """Creates several plasmas with different configurations using the
        Plasma parametric component and checks the location of the x point for
        each."""

        for (
            triangularity,
            elongation,
            minor_radius,
            major_radius,
            vertical_displacement,
        ) in zip(
            [-0.7, 0, 0.5],  # triangularity
            [1, 1.5, 2],  # elongation
            [100, 200, 300],  # minor radius
            [300, 400, 600],  # major radius
            [0, -10, 5],
        ):  # displacement

            for config in ["non-null", "single-null", "double-null"]:

                # Run
                test_plasma = paramak.Plasma(
                    configuration=config,
                    triangularity=triangularity,
                    elongation=elongation,
                    minor_radius=minor_radius,
                    major_radius=major_radius,
                    vertical_displacement=vertical_displacement,
                )

                # Expected
                expected_lower_x_point, expected_upper_x_point = None, None
                if config == "single-null" or config == "double-null":
                    expected_lower_x_point = (1 -
                                              (1 +
                                               test_plasma.x_point_shift) *
                                              triangularity *
                                              minor_radius, -
                                              (1 +
                                               test_plasma.x_point_shift) *
                                              elongation *
                                              minor_radius +
                                              vertical_displacement, )

                    if config == "double-null":
                        expected_upper_x_point = (
                            expected_lower_x_point[0],
                            (1 +
                             test_plasma.x_point_shift) *
                            elongation *
                            minor_radius +
                            vertical_displacement,
                        )

                # Check
                for point, expected_point in zip(
                    [test_plasma.lower_x_point, test_plasma.upper_x_point],
                    [expected_lower_x_point, expected_upper_x_point],
                ):
                    assert point == expected_point

    def test_plasma_x_points_plasmaboundaries(self):
        """Creates several plasmas with different configurations using the
        PlasmaBoundaries parametric component and checks the location of the x
        point for each."""

        for A, triangularity, elongation, minor_radius, major_radius in zip(
            [0, 0.05, 0.05],  # A
            [-0.7, 0, 0.5],  # triangularity
            [1, 1.5, 2],  # elongation
            [100, 200, 300],  # minor radius
            [300, 400, 600],
        ):  # major radius

            for config in ["non-null", "single-null", "double-null"]:

                # Run
                test_plasma = paramak.PlasmaBoundaries(
                    configuration=config,
                    A=A,
                    triangularity=triangularity,
                    elongation=elongation,
                    minor_radius=minor_radius,
                    major_radius=major_radius,
                )

                # Expected
                expected_lower_x_point, expected_upper_x_point = None, None
                if config == "single-null" or config == "double-null":
                    expected_lower_x_point = (1 -
                                              (1 +
                                               test_plasma.x_point_shift) *
                                              triangularity *
                                              minor_radius, -
                                              (1 +
                                               test_plasma.x_point_shift) *
                                              elongation *
                                              minor_radius, )

                    if config == "double-null":
                        expected_upper_x_point = (
                            expected_lower_x_point[0],
                            -expected_lower_x_point[1],
                        )

                # Check
                for point, expected_point in zip(
                    [test_plasma.lower_x_point, test_plasma.upper_x_point],
                    [expected_lower_x_point, expected_upper_x_point],
                ):
                    assert point == expected_point

    def test_plasmaboundaries_solid(self):
        """Create a default PlasmaBoundaries shape and check a solid can be
        created"""
        test_plasma = paramak.PlasmaBoundaries()
        for config in ["non-null", "single-null", "double-null"]:
            test_plasma.configuration = config
            assert test_plasma.solid is not None

    def test_export_plasma_source(self):
        """Creates a plasma using the Plasma parametric component and checks a
        stp file of the shape can be exported using the export_stp method."""

        test_plasma = paramak.Plasma()

        os.system("rm plasma.stp")

        test_plasma.export_stp("plasma.stp")

        assert Path("plasma.stp").exists()
        os.system("rm plasma.stp")

    def test_export_plasma_from_points_export(self):
        """Creates a plasma using the PlasmaFromPoints parametric component
        and checks a stp file of the shape can be exported using the export_stp
        method."""

        test_plasma = paramak.PlasmaFromPoints(
            outer_equatorial_x_point=500,
            inner_equatorial_x_point=300,
            high_point=(400, 200),
            rotation_angle=180,
        )

        os.system("rm plasma.stp")

        test_plasma.export_stp("plasma.stp")
        assert test_plasma.high_point[0] > test_plasma.inner_equatorial_x_point
        assert test_plasma.high_point[0] < test_plasma.outer_equatorial_x_point
        assert test_plasma.outer_equatorial_x_point > test_plasma.inner_equatorial_x_point
        assert Path("plasma.stp").exists()
        os.system("rm plasma.stp")

    def test_plasma_relative_volume(self):
        """Creates plasmas using the Plasma parametric component and checks that
        the relative volumes of the solids created are correct"""

        test_plasma = paramak.Plasma()
        test_plasma_volume = test_plasma.volume()
        test_plasma.rotation_angle = 180
        assert test_plasma.volume() == pytest.approx(test_plasma_volume * 0.5)




from paramak import PoloidalFieldCoil


class PoloidalFieldCoilFP(PoloidalFieldCoil):
    """Creates a rectangular poloidal field coil.

    Args:
        corner_points (list of float tuples): the coordinates of the opposite
            corners of the rectangular shaped coil e.g [(x1, y1), (x2, y2)]
    """

    def __init__(
        self,
        corner_points,
        **kwargs
    ):

        height = abs(corner_points[0][1] - corner_points[1][1])
        width = abs(corner_points[0][0] - corner_points[1][0])

        center_width = (corner_points[0][1] + corner_points[1][1]) / 2.
        center_height = (corner_points[0][1] + corner_points[1][1]) / 2.
        center_point = (center_width, center_height)

        super().__init__(
            height=height,
            width=width,
            center_point=center_point,
            **kwargs
        )

        self.corner_points = corner_points

    @property
    def corner_points(self):
        return self._corner_points

    @corner_points.setter
    def corner_points(self, value):
        # ToDo check the corner points are a list with two entries
        # and each entry is a tuple with two floats
        self._corner_points = value




import os
import unittest
from pathlib import Path

import paramak


class TestTFCoilCasing(unittest.TestCase):

    def setUp(self):
        inner_offset = 10
        outer_offset = 10
        magnet_thickness = 5
        magnet_extrude_distance = 10
        vertical_section_offset = 20
        casing_extrude_distance = magnet_extrude_distance * 2

        # create a princeton D magnet
        magnet = paramak.ToroidalFieldCoilPrincetonD(
            R1=100, R2=200, thickness=magnet_thickness,
            distance=magnet_extrude_distance, number_of_coils=1)

        self.test_shape = paramak.TFCoilCasing(
            magnet=magnet, distance=casing_extrude_distance,
            inner_offset=inner_offset, outer_offset=outer_offset,
            vertical_section_offset=vertical_section_offset)

    def test_creation(self):

        assert self.test_shape.solid is not None

    def test_export_stp(self):
        """Exports and stp file with mode = solid and wire and checks
        that the outputs exist and relative file sizes are correct."""

        os.system("rm test_solid.stp test_solid2.stp test_wire.stp")

        self.test_shape.export_stp('test_solid.stp', mode='solid')
        self.test_shape.export_stp('test_solid2.stp')
        self.test_shape.export_stp('test_wire.stp', mode='wire')

        assert Path("test_solid.stp").exists() is True
        assert Path("test_solid2.stp").exists() is True
        assert Path("test_wire.stp").exists() is True

        assert Path("test_solid.stp").stat().st_size == \
            Path("test_solid2.stp").stat().st_size
        assert Path("test_wire.stp").stat().st_size < \
            Path("test_solid2.stp").stat().st_size

        os.system("rm test_solid.stp test_solid2.stp test_wire.stp")




import unittest

import paramak
import pytest


class TestToroidalFieldCoilRectangle(unittest.TestCase):

    def setUp(self):
        self.test_shape = paramak.ToroidalFieldCoilRectangle(
            horizontal_start_point=(100, 700),
            vertical_mid_point=(800, 0),
            thickness=50,
            distance=30,
            number_of_coils=1,
            with_inner_leg=True
        )

    def test_default_parameters(self):
        """Checks that the default parameters of a ToroidalFieldCoilRectangle are correct."""

        assert self.test_shape.rotation_angle == 360
        assert self.test_shape.with_inner_leg

    def test_points_calculation(self):
        """Checks that the points used to construct the ToroidalFieldCoilRectangle are
        calculated correctly from the parameters given."""

        assert self.test_shape.points == [
            (100, 700),
            (150, 700),
            (800, 700),
            (800, -700),
            (150, -700),
            (100, -700),
            (100, -750),
            (850, -750),
            (850, 750),
            (100, 750)
        ]

    def test_processed_points_calculation(self):
        """Checks that the processed points used to construct the ToroidalFieldCoilRectangle are
        calculated correctly from the parameters given."""

        assert self.test_shape.processed_points == [
            (100, 700, 'straight'),
            (150, 700, 'straight'),
            (800, 700, 'straight'),
            (800, -700, 'straight'),
            (150, -700, 'straight'),
            (100, -700, 'straight'),
            (100, -750, 'straight'),
            (850, -750, 'straight'),
            (850, 750, 'straight'),
            (100, 750, 'straight'),
            (100, 700, 'straight')
        ]

    def test_creation_with_inner_leg(self):
        """Creates a tf coil with inner leg using the ToroidalFieldCoilRectangle
        parametric component and checks that a cadquery solid is created."""

        assert self.test_shape.solid is not None
        assert self.test_shape.volume() > 1000
        assert self.test_shape.inner_leg_connection_points is not None

        test_inner_leg = paramak.ExtrudeStraightShape(
            points=self.test_shape.inner_leg_connection_points, distance=30
        )
        assert test_inner_leg.solid is not None

    def test_creation_with_inner_leg_with_overlap(self):
        """Creates tf coils with overlapping inner legs using the ToroidalFieldCoilRectangle
        parametric component and checks that a cadquery solid is created correctly."""

        self.test_shape.horizontal_start_point = (0, 700)
        self.test_shape.number_of_coils = 8
        assert self.test_shape.solid is not None

        with_inner_leg_volume = self.test_shape.volume()

        self.test_shape.with_inner_leg = False
        without_inner_leg_volume = self.test_shape.volume()

        assert self.test_shape.solid is not None

        test_inner_leg = paramak.ExtrudeStraightShape(
            points=[
                (0, 700), (50, 700), (50, -700), (0, -700)
            ], distance=30,
            azimuth_placement_angle=[0, 45, 90, 135, 180, 225, 270, 315]
        )
        inner_leg_volume = test_inner_leg.volume()

        assert with_inner_leg_volume == pytest.approx(
            without_inner_leg_volume + inner_leg_volume)

    def test_creation_no_inner_leg(self):
        """Creates a tf coil with no inner leg using the ToroidalFieldCoilRectangle
        parametric component and checks that a cadquery solid is created."""

        test_volume = self.test_shape.volume()

        test_inner_leg = paramak.ExtrudeStraightShape(
            points=self.test_shape.inner_leg_connection_points, distance=30
        )
        inner_leg_volume = test_inner_leg.volume()

        self.test_shape.with_inner_leg = False
        assert self.test_shape.solid is not None
        assert self.test_shape.volume() == pytest.approx(
            test_volume - inner_leg_volume)

    def test_absolute_volume(self):
        """Creates a tf coil using the ToroidalFieldCoilRectangle parametric
        component and checks that the volume is correct."""

        self.test_shape.thickness = 150
        self.test_shape.distance = 50

        assert self.test_shape.volume() == pytest.approx(
            (850 * 150 * 50 * 2) + (1400 * 150 * 50 * 2))

        self.test_shape.with_inner_leg = False
        assert self.test_shape.volume() == pytest.approx(
            (850 * 150 * 50 * 2) + (1400 * 150 * 50))

        self.test_shape.with_inner_leg = True
        self.test_shape.number_of_coils = 8
        assert self.test_shape.volume() == pytest.approx(
            ((850 * 150 * 50 * 2) + (1400 * 150 * 50 * 2)) * 8
        )

        self.test_shape.with_inner_leg = False
        assert self.test_shape.volume() == pytest.approx(
            ((850 * 150 * 50 * 2) + (1400 * 150 * 50)) * 8
        )

    def test_absolute_areas(self):
        """Creates tf coils using the ToroidalFieldCoilRectangle parametric
        component and checks that the areas of the faces are correct."""

        self.test_shape.thickness = 150
        self.test_shape.distance = 50

        assert self.test_shape.area == pytest.approx((((850 * 150 * 2) + (1400 * 150)) * 2) + (
            1400 * 150 * 2) + (850 * 50 * 2) + (1700 * 50) + (1400 * 50 * 3) + (700 * 50 * 2) + (150 * 50 * 4))
        assert len(self.test_shape.areas) == 16
        assert self.test_shape.areas.count(pytest.approx(
            (850 * 150 * 2) + (1400 * 150))) == 2
        assert self.test_shape.areas.count(pytest.approx((1400 * 150))) == 2
        assert self.test_shape.areas.count(pytest.approx(850 * 50)) == 2
        assert self.test_shape.areas.count(pytest.approx(1700 * 50)) == 1
        assert self.test_shape.areas.count(pytest.approx(1400 * 50)) == 3
        assert self.test_shape.areas.count(pytest.approx(700 * 50)) == 2
        assert self.test_shape.areas.count(pytest.approx(150 * 50)) == 4

        self.test_shape.with_inner_leg = False
        assert self.test_shape.area == pytest.approx((((850 * 150 * 2) + (1400 * 150)) * 2) + (
            850 * 50 * 2) + (1700 * 50) + (1400 * 50) + (700 * 50 * 2) + (150 * 50 * 2))

    def test_rotation_angle(self):
        """Creates tf coils with rotation_angles < 360 degrees in different
        workplanes and checks that the correct cuts are performed and their
        volumes are correct."""

        self.test_shape.number_of_coils = 8

        self.test_shape.rotation_angle = 360
        self.test_shape.workplane = "XZ"
        test_volume = self.test_shape.volume()
        self.test_shape.rotation_angle = 180
        assert self.test_shape.volume() == pytest.approx(test_volume * 0.5)

        self.test_shape.rotation_angle = 360
        self.test_shape.workplane = "YZ"
        test_volume = self.test_shape.volume()
        self.test_shape.rotation_angle = 180
        assert self.test_shape.volume() == pytest.approx(test_volume * 0.5)

        self.test_shape.rotation_angle = 360
        self.test_shape.workplane = "XY"
        self.test_shape.rotation_axis = "Y"
        test_volume = self.test_shape.volume()
        self.test_shape.rotation_angle = 180
        assert self.test_shape.volume() == pytest.approx(test_volume * 0.5)

    def test_ToroidalFieldCoilRectangle_incorrect_horizonal_start_point(self):
        """Checks that an error is raised when a ToroidalFieldCoilRectangle is made
        with an incorrect horizontal_start_point."""

        def make_ToroidalFieldCoilRectangle_incorrect_horizontal_start_point():
            self.test_shape.vertical_mid_point = (800, 0)
            self.test_shape.horizontal_start_point = (801, 700)
            self.test_shape.solid

        self.assertRaises(
            ValueError,
            make_ToroidalFieldCoilRectangle_incorrect_horizontal_start_point
        )

    def test_ToroidalFieldCoilRectangle_incorrect_vertical_mid_point(self):
        """Checks that an error is raised when a ToroidalFieldCoilRectangle is made
        with an incorrect vertical_mid_point."""

        def make_ToroidalFieldCoilRectangle_incorrect_vertical_mid_point():
            self.test_shape.horizontal_start_point = (100, 700)
            self.test_shape.vertical_mid_point = (800, 701)
            self.test_shape.solid

        self.assertRaises(
            ValueError,
            make_ToroidalFieldCoilRectangle_incorrect_vertical_mid_point
        )




from typing import Optional
import numpy as np
from paramak import SweepCircleShape


class CoolantChannelRingCurved(SweepCircleShape):
    """A ring of equally-spaced curved circular coolant channels with
    constant thickness.

    Args:
        height: height of each coolant channel in ring.
        channel_radius: radius of each coolant channel in ring.
        number_of_coolant_channels: number of coolant channels in ring.
        ring_radius: radius of coolant channel ring.
        start_angle: angle at which the first channel in the ring is placed.
            Defaults to 0.0.
        rotation_axis: azimuthal axis around which the separate coolant channels
            are placed. Default calculated by workplane and path_workplane.
        workplane (str, optional): plane in which the cross-sections of the
            coolant channels lie. Defaults to "XY".
        path_workplane (str, optional): plane in which the cross-sections of the
            coolant channels are swept. Defaults to "XZ".
        force_cross_section (bool, optional): forces coolant channels to have a
            more constant cross-section along their curve. Defaults to False.
    """

    def __init__(
        self,
        height: float,
        channel_radius: float,
        number_of_coolant_channels: int,
        ring_radius: float,
        mid_offset: float,
        start_angle: Optional[float] = 0.0,
        **kwargs
    ) -> None:

        self.ring_radius = ring_radius
        self.mid_offset = mid_offset
        self.height = height
        self.start_angle = start_angle

        super().__init__(
            path_points=self.path_points,
            radius=channel_radius,
            **kwargs
        )

        self.channel_radius = channel_radius
        self.number_of_coolant_channels = number_of_coolant_channels

    @property
    def azimuth_placement_angle(self):
        self.find_azimuth_placement_angle()
        return self._azimuth_placement_angle

    @azimuth_placement_angle.setter
    def azimuth_placement_angle(self, value):
        self._azimuth_placement_angle = value

    @property
    def path_points(self):
        self.find_path_points()
        return self._path_points

    @path_points.setter
    def path_points(self, value):
        self._path_points = value

    def find_azimuth_placement_angle(self):
        """Calculates the azimuth placement angles based on the number of
        coolant channels."""

        angles = list(
            np.linspace(
                0 + self.start_angle,
                360 + self.start_angle,
                self.number_of_coolant_channels,
                endpoint=False))

        self.azimuth_placement_angle = angles

    def find_path_points(self):

        path_points = [
            (self.ring_radius, -self.height / 2),
            (self.ring_radius + self.mid_offset, 0),
            (self.ring_radius, self.height / 2)
        ]

        self.path_points = path_points




import math

from typing import Optional, Tuple
import cadquery as cq
import numpy as np
from paramak import ExtrudeStraightShape
from paramak.utils import calculate_wedge_cut, rotate


class ToroidalFieldCoilCoatHanger(ExtrudeStraightShape):
    """Creates a coat hanger shaped toroidal field coil.

    Args:
        horizontal_start_point: the (x,z) coordinates of the inner upper
            point (cm).
        horizontal_length: the radial length of the horizontal section of
            the TF coil (cm).
        vertical_mid_point: the (x,z) coordinates of the mid point of the
            outboard vertical section (cm).
        vertical_length: the radial length of the outboard vertical section
            of the TF coil (cm).
        thickness: the thickness of the toroidal field coil.
        distance: the extrusion distance.
        number_of_coils: the number of TF coils. This changes with
            azimuth_placement_angle dividing up 360 degrees by the number of
            coils.
        with_inner_leg: Include the inner TF leg. Defaults to True.
    """

    def __init__(
        self,
        horizontal_start_point: Tuple[float, float],
        horizontal_length: float,
        vertical_mid_point: Tuple[float, float],
        vertical_length: float,
        thickness: float,
        distance: float,
        number_of_coils: int,
        with_inner_leg: Optional[bool] = True,
        color: Optional[Tuple[float, float, float, Optional[float]]] = (0., 0., 1.),
        **kwargs
    ) -> None:

        super().__init__(
            distance=distance,
            color=color,
            **kwargs
        )

        self.horizontal_start_point = horizontal_start_point
        self.horizontal_length = horizontal_length
        self.vertical_mid_point = vertical_mid_point
        self.vertical_length = vertical_length
        self.thickness = thickness
        self.distance = distance
        self.number_of_coils = number_of_coils
        self.with_inner_leg = with_inner_leg

    @property
    def azimuth_placement_angle(self):
        self.find_azimuth_placement_angle()
        return self._azimuth_placement_angle

    @azimuth_placement_angle.setter
    def azimuth_placement_angle(self, value):
        self._azimuth_placement_angle = value

    def find_points(self):
        """Finds the XZ points joined by straight connections that describe the
        2D profile of the poloidal field coil shape."""

        # 16---15
        # -     -
        # -       14
        # -        -
        # 1---2     -
        #       -    -
        #        -    13
        #         -    -
        #          3    12
        #          -    -
        #          -    -
        #          -    -
        #          4    11
        #         -    -
        #        -    10
        #       -    -
        # 6---5     -
        # -       -
        # -      9
        # -    -
        # 7---8

        adjacent_length = self.vertical_mid_point[0] - (
            self.horizontal_start_point[0] + self.horizontal_length)
        oppersite_length = self.horizontal_start_point[1] - (
            self.vertical_mid_point[1] + 0.5 * self.vertical_length)

        point_rotation = math.atan(oppersite_length / adjacent_length)
        point_rotation_mid = math.radians(90) - point_rotation

        points = [
            self.horizontal_start_point,  # point 1
            (
                self.horizontal_start_point[0] + self.horizontal_length,
                self.horizontal_start_point[1],
            ),  # point 2
            (
                self.vertical_mid_point[0],
                self.vertical_mid_point[1] + 0.5 * self.vertical_length,
            ),  # point 3
            (
                self.vertical_mid_point[0],
                self.vertical_mid_point[1] - 0.5 * self.vertical_length,
            ),  # point 4
            (
                self.horizontal_start_point[0] + self.horizontal_length,
                -self.horizontal_start_point[1],
            ),  # point 5
            (
                self.horizontal_start_point[0],
                -self.horizontal_start_point[1],
            ),  # point 6
            (
                self.horizontal_start_point[0],
                -self.horizontal_start_point[1] - self.thickness,
            ),  # point 7
            (
                self.horizontal_start_point[0] + self.horizontal_length,
                -self.horizontal_start_point[1] - self.thickness,
            ),  # point 8
            rotate(
                (
                    self.horizontal_start_point[0] + self.horizontal_length,
                    -self.horizontal_start_point[1],
                ),  # same as point 5
                (
                    self.horizontal_start_point[0] + self.horizontal_length,
                    -self.horizontal_start_point[1] - self.thickness,
                ),  # same as point 8
                point_rotation
            ),  # point 9
            rotate(
                (
                    self.vertical_mid_point[0],
                    self.vertical_mid_point[1] - 0.5 * self.vertical_length,
                ),  # same as point 4
                (
                    self.vertical_mid_point[0] + self.thickness,
                    self.vertical_mid_point[1] - 0.5 * self.vertical_length,
                ),  # same as point 11
                -point_rotation_mid
            ),  # point 10
            (
                self.vertical_mid_point[0] + self.thickness,
                self.vertical_mid_point[1] - 0.5 * self.vertical_length,
            ),  # point 11
            (
                self.vertical_mid_point[0] + self.thickness,
                self.vertical_mid_point[1] + 0.5 * self.vertical_length,
            ),  # point 12
            rotate(
                (
                    self.vertical_mid_point[0],
                    self.vertical_mid_point[1] + 0.5 * self.vertical_length,
                ),  # same as point 3
                (
                    self.vertical_mid_point[0] + self.thickness,
                    self.vertical_mid_point[1] + 0.5 * self.vertical_length,
                ),  # same as point 12
                point_rotation_mid
            ),  # point 13
            rotate(
                (
                    self.horizontal_start_point[0] + self.horizontal_length,
                    self.horizontal_start_point[1],
                ),  # same as point 2
                (
                    self.horizontal_start_point[0] + self.horizontal_length,
                    self.horizontal_start_point[1] + self.thickness,
                ),  # same as point 15
                -point_rotation
            ),  # point 14
            (
                self.horizontal_start_point[0] + self.horizontal_length,
                self.horizontal_start_point[1] + self.thickness,
            ),  # point 15
            (
                self.horizontal_start_point[0],
                self.horizontal_start_point[1] + self.thickness,
            )   # point 16
        ]

        self.inner_leg_connection_points = [
            points[0],
            (points[0][0] + self.thickness, points[0][1]),
            (points[5][0] + self.thickness, points[5][1]),
            points[5],
        ]

        self.points = points

    def find_azimuth_placement_angle(self):
        """Calculates the azimuth placement angles based on the number of
        toroidal field coils"""

        angles = list(
            np.linspace(
                0,
                360,
                self.number_of_coils,
                endpoint=False))

        self.azimuth_placement_angle = angles

    def create_solid(self):
        """Creates a 3d solid using points with straight edges.

        Returns:
           A CadQuery solid: A 3D solid volume
        """

        # Creates a cadquery solid from points and revolves
        points_without_connections = [p[:2] for p in self.points]
        solid = (
            cq.Workplane(self.workplane)
            .polyline(points_without_connections)
        )

        wire = solid.close()

        self.wire = wire

        solid = wire.extrude(distance=-self.distance / 2.0, both=True)

        solid = self.rotate_solid(solid)

        cutting_wedge = calculate_wedge_cut(self)
        solid = self.perform_boolean_operations(solid, wedge_cut=cutting_wedge)

        if self.with_inner_leg is True:
            inner_leg_solid = cq.Workplane(self.workplane)
            inner_leg_solid = inner_leg_solid.polyline(
                self.inner_leg_connection_points)
            inner_leg_solid = inner_leg_solid.close().extrude(
                distance=-self.distance / 2.0, both=True)

            inner_leg_solid = self.rotate_solid(inner_leg_solid)
            inner_leg_solid = self.perform_boolean_operations(
                inner_leg_solid, wedge_cut=cutting_wedge)

            solid = cq.Compound.makeCompound(
                [a.val() for a in [inner_leg_solid, solid]]
            )

        self.solid = solid   # not necessarily required as set in boolean_operations

        return solid



from .shape import Shape
from .reactor import Reactor
from .utils import rotate, extend, distance_between_two_points, diff_between_angles
from .utils import EdgeLengthSelector, FaceAreaSelector

from .parametric_shapes.extruded_mixed_shape import ExtrudeMixedShape
from .parametric_shapes.extruded_spline_shape import ExtrudeSplineShape
from .parametric_shapes.extruded_straight_shape import ExtrudeStraightShape
from .parametric_shapes.extruded_circle_shape import ExtrudeCircleShape

from .parametric_shapes.rotate_mixed_shape import RotateMixedShape
from .parametric_shapes.rotate_spline_shape import RotateSplineShape
from .parametric_shapes.rotate_straight_shape import RotateStraightShape
from .parametric_shapes.rotate_circle_shape import RotateCircleShape

from .parametric_shapes.sweep_mixed_shape import SweepMixedShape
from .parametric_shapes.sweep_spline_shape import SweepSplineShape
from .parametric_shapes.sweep_straight_shape import SweepStraightShape
from .parametric_shapes.sweep_circle_shape import SweepCircleShape

from .parametric_components.hexagon_pin import HexagonPin

from .parametric_components.tokamak_plasma import Plasma
from .parametric_components.tokamak_plasma_from_points import PlasmaFromPoints
from .parametric_components.tokamak_plasma_plasmaboundaries import PlasmaBoundaries

from .parametric_components.blanket_constant_thickness_arc_h import BlanketConstantThicknessArcH
from .parametric_components.blanket_constant_thickness_arc_v import BlanketConstantThicknessArcV
from .parametric_components.blanket_fp import BlanketFP
from .parametric_components.blanket_poloidal_segment import BlanketFPPoloidalSegments

from .parametric_components.divertor_ITER import ITERtypeDivertor
from .parametric_components.divertor_ITER_no_dome import ITERtypeDivertorNoDome

from .parametric_components.center_column_cylinder import CenterColumnShieldCylinder
from .parametric_components.center_column_hyperbola import CenterColumnShieldHyperbola
from .parametric_components.center_column_flat_top_hyperbola import CenterColumnShieldFlatTopHyperbola
from .parametric_components.center_column_plasma_dependant import CenterColumnShieldPlasmaHyperbola
from .parametric_components.center_column_circular import CenterColumnShieldCircular
from .parametric_components.center_column_flat_top_circular import CenterColumnShieldFlatTopCircular

from .parametric_components.coolant_channel_ring_straight import CoolantChannelRingStraight
from .parametric_components.coolant_channel_ring_curved import CoolantChannelRingCurved

from .parametric_components.inboard_firstwall_fccs import InboardFirstwallFCCS

from .parametric_components.poloidal_field_coil import PoloidalFieldCoil
from .parametric_components.poloidal_field_coil_fp import PoloidalFieldCoilFP
from .parametric_components.poloidal_field_coil_case import PoloidalFieldCoilCase
from .parametric_components.poloidal_field_coil_case_fc import PoloidalFieldCoilCaseFC
from .parametric_components.poloidal_field_coil_set import PoloidalFieldCoilSet
from .parametric_components.poloidal_field_coil_case_set import PoloidalFieldCoilCaseSet
from .parametric_components.poloidal_field_coil_case_set_fc import PoloidalFieldCoilCaseSetFC
from .parametric_components.extrude_rectangle import ExtrudeRectangle

from .parametric_components.poloidal_segmenter import PoloidalSegments
from .parametric_components.port_cutters_rotated import PortCutterRotated
from .parametric_components.port_cutters_rectangular import PortCutterRectangular
from .parametric_components.port_cutters_circular import PortCutterCircular
from .parametric_components.rotated_trapezoid import RotatedTrapezoid
from .parametric_components.rotated_isosceles_triangle import RotatedIsoscelesTriangle
from .parametric_components.cutting_wedge import CuttingWedge
from .parametric_components.cutting_wedge_fs import CuttingWedgeFS
from .parametric_components.blanket_cutter_parallels import BlanketCutterParallels
from .parametric_components.blanket_cutters_star import BlanketCutterStar

from .parametric_components.inner_tf_coils_circular import InnerTfCoilsCircular
from .parametric_components.inner_tf_coils_flat import InnerTfCoilsFlat

from .parametric_components.toroidal_field_coil_coat_hanger import ToroidalFieldCoilCoatHanger
from .parametric_components.toroidal_field_coil_rectangle import ToroidalFieldCoilRectangle
from .parametric_components.toroidal_field_coil_triple_arc import ToroidalFieldCoilTripleArc
from .parametric_components.toroidal_field_coil_princeton_d import ToroidalFieldCoilPrincetonD
from .parametric_components.tf_coil_casing import TFCoilCasing
from .parametric_components.toroidal_field_coil_round_corners \
    import ToroidalFieldCoilRectangleRoundCorners

from .parametric_components.vacuum_vessel import VacuumVessel
from .parametric_components.vacuum_vessel_inner_leg import VacuumVesselInnerLeg
from .parametric_components.capsule_vacuum_vessel import CapsuleVacuumVessel
from .parametric_components.hollow_cube import HollowCube
from .parametric_components.shell_fs import ShellFS

from .parametric_reactors.eu_demo_2015_reactor import EuDemoFrom2015PaperDiagram
from .parametric_reactors.ball_reactor import BallReactor
from .parametric_reactors.submersion_reactor import SubmersionTokamak
from .parametric_reactors.single_null_submersion_reactor import SingleNullSubmersionTokamak
from .parametric_reactors.single_null_ball_reactor import SingleNullBallReactor
from .parametric_reactors.segmented_blanket_ball_reactor import SegmentedBlanketBallReactor
from .parametric_reactors.center_column_study_reactor import CenterColumnStudyReactor
from .parametric_reactors.sparc_paper_2020 import SparcFrom2020PaperDiagram
from .parametric_reactors.iter_paper_2020 import IterFrom2020PaperDiagram
from .parametric_reactors.cylinder_reactor import CylinderReactor




import os
import sys
from pathlib import Path
import time
from notebook_testing import notebook_run

sys.path.append(os.path.join(os.path.dirname(__file__), '..', 'examples'))


def main():
    timings = []
    for notebook in Path().rglob("examples/example_parametric_reactors/*.ipynb"):
        start = time.time()

        print(notebook)
        errors = notebook_run(notebook)
        assert errors == []

        stop = time.time()
        duration = stop - start

        print((notebook, duration))
        timings.append((notebook, duration))

    # to see timings run with pytest --capture=tee-sys
    print(timings)


if __name__ == "__main__":
    main()




import unittest

import numpy as np
import paramak


class TestInboardFirstwallFCCS(unittest.TestCase):

    def test_construction_with_CenterColumnShieldCylinder(self):
        """Makes a firstwall from a CenterColumnShieldCylinder and checks
        the volume is smaller than the shield."""

        test_shape_1 = paramak.CenterColumnShieldCylinder(
            height=100,
            inner_radius=20,
            outer_radius=80)
        test_shape_2 = paramak.InboardFirstwallFCCS(
            central_column_shield=test_shape_1,
            thickness=20,
            rotation_angle=180)
        assert test_shape_1.solid is not None
        assert test_shape_2.solid is not None
        assert test_shape_1.volume() > test_shape_2.volume()

    def test_construction_with_CenterColumnShieldHyperbola(self):
        """Makes a firstwall from a CenterColumnShieldHyperbola and checks
        the volume is smaller than the shield."""

        a = paramak.CenterColumnShieldHyperbola(
            height=200,
            inner_radius=20,
            mid_radius=80,
            outer_radius=120)
        b = paramak.InboardFirstwallFCCS(
            central_column_shield=a,
            thickness=20,
            rotation_angle=180)
        assert a.solid is not None
        assert b.solid is not None
        assert a.volume() > b.volume()

    def test_construction_with_CenterColumnShieldFlatTopHyperbola(self):
        """Makes a firstwall from a CenterColumnShieldFlatTopHyperbola and
        checks the volume is smaller than the shield."""

        a = paramak.CenterColumnShieldFlatTopHyperbola(
            height=200,
            arc_height=100,
            inner_radius=50,
            mid_radius=80,
            outer_radius=100)
        b = paramak.InboardFirstwallFCCS(
            central_column_shield=a,
            thickness=20,
            rotation_angle=180)
        assert a.solid is not None
        assert b.solid is not None
        assert a.volume() > b.volume()

    def test_construction_with_CenterColumnShieldPlasmaHyperbola(self):
        """Makes a firstwall from a CenterColumnShieldPlasmaHyperbola and
        checks the volume is smaller than the shield."""

        a = paramak.CenterColumnShieldPlasmaHyperbola(
            height=601,
            inner_radius=20,
            mid_offset=50,
            edge_offset=0)
        b = paramak.InboardFirstwallFCCS(
            central_column_shield=a,
            thickness=20,
            rotation_angle=180)
        assert a.solid is not None
        assert b.solid is not None
        assert a.volume() > b.volume()

    def test_construction_with_CenterColumnShieldCircular(self):
        """Makes a firstwall from a CenterColumnShieldCircular and checks
        the volume is smaller than the shield."""

        a = paramak.CenterColumnShieldCircular(
            height=300,
            inner_radius=20,
            mid_radius=50,
            outer_radius=100)
        b = paramak.InboardFirstwallFCCS(
            central_column_shield=a,
            thickness=20,
            rotation_angle=180)
        assert a.solid is not None
        assert b.solid is not None
        assert a.volume() > b.volume()

    def test_construction_with_CenterColumnShieldFlatTopCircular(self):
        """Makes a firstwall from a CenterColumnShieldFlatTopCircular and
        checks the volume is smaller than the shield."""

        a = paramak.CenterColumnShieldFlatTopCircular(
            height=500,
            arc_height=300,
            inner_radius=30,
            mid_radius=70,
            outer_radius=120)
        b = paramak.InboardFirstwallFCCS(
            central_column_shield=a,
            thickness=20,
            rotation_angle=180)
        assert a.solid is not None
        assert b.solid is not None
        assert a.volume() > b.volume()

    def test_construction_with_wrong_column_shield_type(self):
        def test_construction_with_string():
            """Only CenterColumnShields are acceptable inputs for inputs, this
            should fail as it trys to use a string."""

            b = paramak.InboardFirstwallFCCS(
                central_column_shield="incorrect type",
                thickness=20,
                rotation_angle=180)
            b.solid

        self.assertRaises(
            ValueError,
            test_construction_with_string)

    def test_boolean_union(self):
        """Makes two halves of a 360 firstwall and performs a union and checks
        that the volume corresponds to 2 times the volume of 1 half."""

        a = paramak.CenterColumnShieldFlatTopCircular(
            height=500,
            arc_height=300,
            inner_radius=30,
            mid_radius=70,
            outer_radius=120)
        b = paramak.InboardFirstwallFCCS(
            central_column_shield=a,
            thickness=20,
            rotation_angle=180,
            azimuth_placement_angle=0)

        c = paramak.InboardFirstwallFCCS(
            central_column_shield=a,
            thickness=20,
            rotation_angle=180,
            azimuth_placement_angle=180,
            union=b)
        assert np.isclose(c.volume(), 2 * b.volume())

    def test_azimuth_placement_angle(self):
        """Makes two 180 degree firstwalls (one is rotated 90 degree), performs
        a cut and checks that the volume corresponds to 0.5 times the volume of
        1 half."""

        a = paramak.CenterColumnShieldFlatTopCircular(
            height=500,
            arc_height=300,
            inner_radius=30,
            mid_radius=70,
            outer_radius=120)
        b = paramak.InboardFirstwallFCCS(
            central_column_shield=a,
            thickness=20,
            rotation_angle=180,
            azimuth_placement_angle=0)

        c = paramak.InboardFirstwallFCCS(
            central_column_shield=a,
            thickness=20,
            rotation_angle=180,
            azimuth_placement_angle=90,
            cut=b)
        assert np.isclose(c.volume(), 0.5 * b.volume())

    def test_cut_attribute(self):
        """Creates a firstwall then resets its cut attribute and checks that
        the shape is not affected."""

        a = paramak.CenterColumnShieldCylinder(
            height=100,
            inner_radius=20,
            outer_radius=80)
        b = paramak.InboardFirstwallFCCS(
            central_column_shield=a,
            thickness=20,
            rotation_angle=180)

        volume_1 = b.volume()
        b.cut = None
        volume_2 = b.volume()
        assert np.isclose(volume_1, volume_2)




from typing import Optional, Tuple
import numpy as np
from paramak import ExtrudeMixedShape


class ToroidalFieldCoilTripleArc(ExtrudeMixedShape):
    """Toroidal field coil made of three arcs

    Args:
        R1: smallest radius (cm).
        h: height of the straight section (cm).
        radii: radii of the small and medium arcs (cm).
        coverages: coverages of the small and medium arcs (deg).
        thickness: magnet thickness (cm).
        distance: extrusion distance (cm).
        number_of_coils: the number of TF coils. This changes by the
            azimuth_placement_angle dividing up 360 degrees by the number of
            coils.
        vertical_displacement: vertical displacement (cm). Defaults to 0.0.
        with_inner_leg: Include the inner tf leg. Defaults to True.
    """

    def __init__(
        self,
        R1: float,
        h: float,
        radii: Tuple[float, float],
        coverages: Tuple[float, float],
        thickness: float,
        distance: float,
        number_of_coils: int,
        vertical_displacement: Optional[float] = 0.0,
        with_inner_leg: Optional[bool] = True,
        color: Optional[Tuple[float, float, float, Optional[float]]] = (0., 0., 1.),
        **kwargs
    ) -> None:

        super().__init__(
            distance=distance,
            color=color,
            **kwargs
        )
        self.R1 = R1
        self.h = h
        self.small_radius, self.mid_radius = radii

        self.small_coverage, self.mid_coverage = coverages
        self.thickness = thickness
        self.distance = distance
        self.number_of_coils = number_of_coils
        self.vertical_displacement = vertical_displacement
        self.with_inner_leg = with_inner_leg

    @property
    def azimuth_placement_angle(self):
        self.find_azimuth_placement_angle()
        return self._azimuth_placement_angle

    @azimuth_placement_angle.setter
    def azimuth_placement_angle(self, value):
        self._azimuth_placement_angle = value

    def _compute_curve(self, R1, h, radii, coverages):
        npoints = 500

        small_radius, mid_radius = radii
        small_coverage, mid_coverage = coverages
        asum = small_coverage + mid_coverage

        # small arc
        theta = np.linspace(
            0, small_coverage, round(0.5 * npoints * small_coverage / np.pi))
        small_arc_R = R1 + small_radius * (1 - np.cos(theta))
        small_arc_Z = h + small_radius * np.sin(theta)

        # mid arc
        theta = np.linspace(
            theta[-1], asum, round(0.5 * npoints * mid_coverage / np.pi))
        mid_arc_R = small_arc_R[-1] + mid_radius * \
            (np.cos(small_coverage) - np.cos(theta))
        mid_arc_Z = small_arc_Z[-1] + mid_radius * \
            (np.sin(theta) - np.sin(small_coverage))

        # large arc
        large_radius = (mid_arc_Z[-1]) / np.sin(np.pi - asum)
        theta = np.linspace(theta[-1], np.pi, 60)
        large_arc_R = mid_arc_R[-1] + large_radius * \
            (np.cos(np.pi - theta) - np.cos(np.pi - asum))
        large_arc_Z = mid_arc_Z[-1] - large_radius * \
            (np.sin(asum) - np.sin(np.pi - theta))

        R = np.concatenate((small_arc_R, mid_arc_R[1:], large_arc_R[1:]))
        R = np.append(R, np.flip(R)[1:])
        Z = np.concatenate((small_arc_Z, mid_arc_Z[1:], large_arc_Z[1:]))
        Z = np.append(Z, -np.flip(Z)[1:])
        return R, Z

    def find_points(self):
        """Finds the XZ points joined by connections that describe the 2D
        profile of the toroidal field coil shape."""

        thickness = self.thickness
        small_radius, mid_radius = self.small_radius, self.mid_radius
        small_coverage, mid_coverage = self.small_coverage, self.mid_coverage
        small_coverage *= np.pi / 180  # convert to radians
        mid_coverage *= np.pi / 180

        # create inner coordinates
        R_inner, Z_inner = self._compute_curve(
            self.R1 + thickness, self.h * 0.5, radii=(small_radius, mid_radius),
            coverages=(small_coverage, mid_coverage))

        # create outer coordinates
        R_outer, Z_outer = self._compute_curve(
            self.R1, self.h * 0.5,
            radii=(small_radius + thickness, mid_radius + thickness),
            coverages=(small_coverage, mid_coverage))
        R_outer, Z_outer = np.flip(R_outer), np.flip(Z_outer)

        # add vertical displacement
        Z_outer += self.vertical_displacement
        Z_inner += self.vertical_displacement

        # extract helping points for inner leg
        inner_leg_connection_points = [
            (R_inner[0], Z_inner[0]),
            (R_inner[-1], Z_inner[-1]),
            (R_outer[0], Z_outer[0]),
            (R_outer[-1], Z_outer[-1])
        ]
        self.inner_leg_connection_points = inner_leg_connection_points

        # add the leg to the points
        if self.with_inner_leg:
            R_inner = np.append(R_inner, R_inner[0])
            Z_inner = np.append(Z_inner, Z_inner[0])

            R_outer = np.append(R_outer, R_outer[0])
            Z_outer = np.append(Z_outer, Z_outer[0])
        # add connections
        inner_points = [[r, z, 'spline'] for r, z in zip(R_inner, Z_inner)]
        outer_points = [[r, z, 'spline'] for r, z in zip(R_outer, Z_outer)]
        if self.with_inner_leg:
            outer_points[-2][2] = 'straight'
            inner_points[-2][2] = 'straight'

        inner_points[-1][2] = 'straight'
        outer_points[-1][2] = 'straight'

        points = inner_points + outer_points

        self.points = points

    def find_azimuth_placement_angle(self):
        """Calculates the azimuth placement angles based on the number of tf
        coils"""

        angles = list(
            np.linspace(
                0,
                360,
                self.number_of_coils,
                endpoint=False))

        self.azimuth_placement_angle = angles




import math
import os
import unittest
from pathlib import Path

import pytest
from paramak import RotateCircleShape


class TestRotateCircleShape(unittest.TestCase):

    def setUp(self):
        self.test_shape = RotateCircleShape(
            points=[(60, 0)],
            radius=10
        )

    def test_default_parameters(self):
        """Checks that the default parameters of a RotateCircleShape are correct."""

        assert self.test_shape.rotation_angle == 360
        assert self.test_shape.azimuth_placement_angle == 0

    def test_absolute_shape_volume(self):
        """Creates RotateCircleShapes and checks that their volumes are correct."""

        # See Issue #445
        # assert self.test_shape.volume == pytest.approx(
        #     2 * math.pi * 60 * math.pi * (10**2)
        # )
        self.test_shape.rotation_angle = 270
        assert self.test_shape.volume() == pytest.approx(
            2 * math.pi * 60 * math.pi * (10**2) * 0.75
        )

    def test_absolute_shape_areas(self):
        """Creates RotateCircleShapes and checks that the areas of each face are
        correct."""

        # See Issue #445
        # assert self.test_shape.area == pytest.approx(
        #     math.pi * (10 * 2) * math.pi * (60 * 2))
        # assert len(self.test_shape.areas) == 1
        # assert self.test_shape.areas.count(pytest.approx(
        #     math.pi * (10 * 2) * math.pi * (60 * 2), rel=0.01)) == 1

        self.test_shape.rotation_angle = 180
        assert self.test_shape.area == pytest.approx(
            ((math.pi * (10**2)) * 2) + (math.pi * (10 * 2) * math.pi * (60 * 2) / 2), rel=0.01)
        assert len(self.test_shape.areas) == 3
        assert self.test_shape.areas.count(
            pytest.approx(math.pi * (10**2))) == 2
        assert self.test_shape.areas.count(pytest.approx(
            math.pi * (10 * 2) * math.pi * (60 * 2) / 2, rel=0.01)) == 1

    def test_relative_shape_volume_azimuth_placement_angle(self):
        """Creates two RotateCircleShapes with different
        azimuth_placement_angles and checks that their relative volumes are
        correct."""

        self.test_shape.rotation_angle = 10
        assert self.test_shape.volume() == pytest.approx(
            (math.pi * 10**2) * ((2 * math.pi * 60) / 36)
        )

        self.test_shape.azimuth_placement_angle = [0, 90, 180, 270]
        assert self.test_shape.volume() == pytest.approx(
            (math.pi * 10**2) * ((2 * math.pi * 60) / 36) * 4
        )

    def test_cut_volume(self):
        """Creates a RotateCircleShape with another RotateCircleShape cut out
        and checks that the volume is correct."""

        outer_shape = RotateCircleShape(
            points=[(60, 0)], radius=15
        )
        outer_shape_volume = outer_shape.volume()
        outer_shape.cut = self.test_shape
        assert outer_shape.volume() == pytest.approx(
            outer_shape_volume - self.test_shape.volume()
        )

    def test_export_stp(self):
        """Exports and stp file with mode = solid and wire and checks
        that the outputs exist and relative file sizes are correct."""

        os.system("rm test_solid.stp test_solid2.stp test_wire.stp")

        self.test_shape.export_stp('test_solid.stp', mode='solid')
        self.test_shape.export_stp('test_solid2.stp')
        self.test_shape.export_stp('test_wire.stp', mode='wire')

        assert Path("test_solid.stp").exists() is True
        assert Path("test_solid2.stp").exists() is True
        assert Path("test_wire.stp").exists() is True

        assert Path("test_solid.stp").stat().st_size == \
            Path("test_solid2.stp").stat().st_size
        # the circle wire file is actually larger than the circle solid file
        # assert Path("test_wire.stp").stat().st_size < \
        #     Path("test_solid2.stp").stat().st_size

        os.system("rm test_solid.stp test_solid2.stp test_wire.stp")


if __name__ == "__main__":
    unittest.main()




import os
import unittest
from pathlib import Path

import pytest
from paramak import RotateMixedShape


class TestRotateMixedShape(unittest.TestCase):

    def setUp(self):
        self.test_shape = RotateMixedShape(
            points=[(50, 0, "straight"), (50, 50, "spline"), (60, 70, "spline"),
                    (70, 50, "circle"), (60, 25, "circle"), (70, 0, "straight")]
        )

        self.test_shape_2 = RotateMixedShape(
            rotation_angle=1,
            points=[
                (100, 0, "straight"),
                (200, 0, "circle"),
                (250, 50, "circle"),
                (200, 100, "straight"),
                (150, 100, "straight"),
                (140, 75, "straight"),
                (110, 45, "straight"),
            ]
        )

        self.test_shape_3 = RotateMixedShape(
            rotation_angle=180,
            points=[
                (100, 0, "straight"),
                (200, 0, "circle"),
                (250, 50, "circle"),
                (200, 100, "straight"),
                (150, 100, "straight"),
                (140, 75, "circle"),
                (110, 45, "circle"),
            ]
        )

    def test_export_2d_image(self):
        """Creates a RotateMixedShape object and checks that a png file of the
        object with the correct suffix can be exported using the
        export_2d_image method."""

        os.system("rm filename.png")
        self.test_shape.export_2d_image("filename")
        assert Path("filename.png").exists() is True
        os.system("rm filename.png")
        self.test_shape.export_2d_image("filename.png")
        assert Path("filename.png").exists() is True
        os.system("rm filename.png")

    def test_default_parameters(self):
        """Checks that the default parameters of a RotateMixedShape are correct."""

        assert self.test_shape.rotation_angle == 360
        assert self.test_shape.azimuth_placement_angle == 0

    def test_relative_shape_volume_rotation_angle(self):
        """Creates two RotateMixedShapes with different rotation_angles and checks
        that their relative volumes are correct."""

        assert self.test_shape.volume() > 100
        test_volume = self.test_shape.volume()
        self.test_shape.rotation_angle = 180
        assert test_volume == pytest.approx(self.test_shape.volume() * 2)

    def test_relative_shape_volume_azimuth_placement_angle(self):
        """Creates two RotateMixedShapes with different azimuth_placement_angles and
        checks that their relative volumes are correct."""

        self.test_shape.rotation_angle = 10
        self.test_shape.azimuth_placement_angle = 0
        test_volume_1 = self.test_shape.volume()

        self.test_shape.azimuth_placement_angle = [0, 90, 180, 270]
        assert self.test_shape.volume() == pytest.approx(test_volume_1 * 4)

    def test_shape_face_areas(self):
        """Creates RotateMixedShapes and checks that the face areas are expected."""

        assert len(self.test_shape.areas) == 4
        assert len(set(self.test_shape.areas)) == 4

        self.test_shape.rotation_angle = 180
        assert len(self.test_shape.areas) == 6
        assert len(set([round(i) for i in self.test_shape.areas])) == 5

    def test_union_volume_addition(self):
        """Fuses two RotateMixedShapes and checks that their fused volume
        is correct"""

        inner_box = RotateMixedShape(
            points=[
                (100, 100, "straight"),
                (100, 200, "straight"),
                (200, 200, "straight"),
                (200, 100, "straight"),
            ]
        )

        outer_box = RotateMixedShape(
            points=[
                (200, 100, "straight"),
                (200, 200, "straight"),
                (500, 200, "straight"),
                (500, 100, "straight"),
            ]
        )

        outer_box_and_inner_box = RotateMixedShape(
            points=[
                (200, 100, "straight"),
                (200, 200, "straight"),
                (500, 200, "straight"),
                (500, 100, "straight"),
            ],
            union=inner_box,
        )

        assert inner_box.volume() + outer_box.volume() == pytest.approx(
            outer_box_and_inner_box.volume()
        )

    def test_incorrect_connections(self):
        """Checks that errors are raised when invalid connection arguments are
        used to construct a RotateMixedShape."""

        def incorrect_string_for_connection_type():
            """Checks ValueError is raised when an invalid connection type is
            specified."""

            RotateMixedShape(
                points=[
                    (0, 0, "straight"),
                    (0, 20, "spline"),
                    (20, 20, "spline"),
                    (20, 0, "invalid_entry"),
                ]
            )

        self.assertRaises(ValueError, incorrect_string_for_connection_type)

        def incorrect_number_of_connections_function():
            """Checks ValueError is raised when an incorrect number of
            connections is specified."""

            RotateMixedShape(
                points=[(0, 200, "straight"), (200, 100), (0, 0, "spline")]
            )

        self.assertRaises(ValueError, incorrect_number_of_connections_function)

    def test_cut_volume(self):
        """Creates a RotateMixedShape with another RotateMixedShape cut out and
        checks that the volume is correct."""

        outer_shape = RotateMixedShape(
            points=[
                (40, -10, "spline"),
                (35, 50, "spline"),
                (60, 80, "straight"),
                (80, 80, "circle"),
                (100, 40, "circle"),
                (80, 0, "straight"),
                (80, -10, "straight")
            ]
        )
        outer_shape_volume = outer_shape.volume()
        outer_shape.cut = self.test_shape
        assert outer_shape.volume() == pytest.approx(
            outer_shape_volume - self.test_shape.volume()
        )

    def test_mixed_shape_with_straight_and_circle(self):
        """Creates a RotateMixedShape with straight and circular connections and
        checks that the volume is correct."""

        test_shape = RotateMixedShape(
            points=[
                (10, 20, "straight"),
                (10, 10, "straight"),
                (20, 10, "circle"),
                (40, 15, "circle"),
                (20, 20, "straight"),
            ],
            rotation_angle=10,
        )
        assert test_shape.volume() > 10 * 10

    def test_export_stp_extension(self):
        """Creates a RotateMixedShape and checks that a stp file of the shape
        can be exported with the correct suffix using the export_stp method."""

        os.system("rm filename.stp filename.step")
        self.test_shape.export_stp("filename.stp")
        self.test_shape.export_stp("filename.step")
        assert Path("filename.stp").exists() is True
        assert Path("filename.step").exists() is True
        os.system("rm filename.stp filename.step")

    def test_export_stl(self):
        """Creates a RotateMixedShape and checks that a stl file of the shape
        can be exported with the correct suffix using the export_stl method."""

        os.system("rm filename.stl")
        self.test_shape.export_stl("filename.stl")
        assert Path("filename.stl").exists() is True
        os.system("rm filename.stl")

    def test_export_stp(self):
        """Exports and stp file with mode = solid and wire and checks
        that the outputs exist and relative file sizes are correct."""

        os.system("rm test_solid.stp test_solid2.stp test_wire.stp")

        self.test_shape.export_stp('test_solid.stp', mode='solid')
        self.test_shape.export_stp('test_solid2.stp')
        self.test_shape.export_stp('test_wire.stp', mode='wire')

        assert Path("test_solid.stp").exists() is True
        assert Path("test_solid2.stp").exists() is True
        assert Path("test_wire.stp").exists() is True

        assert Path("test_solid.stp").stat().st_size == \
            Path("test_solid2.stp").stat().st_size
        assert Path("test_wire.stp").stat().st_size < \
            Path("test_solid2.stp").stat().st_size

        os.system("rm test_solid.stp test_solid2.stp test_wire.stp")

    def test_convert_all_circle_points_change_to_splines(self):
        """creates a RotateMixedShape with two circular edges and converts
        them to spline edges. Checks the new edges have been correctly
        replaced with splines"""

        assert len(self.test_shape_3.processed_points) == 8
        self.test_shape_3.convert_all_circle_connections_to_splines()
        assert len(self.test_shape_3.processed_points) > 8
        assert self.test_shape_3.processed_points[0] == (100, 0, "straight")
        assert self.test_shape_3.processed_points[1][2] == 'spline'
        assert self.test_shape_3.processed_points[2][2] == 'spline'

        # last point is the same as the first point
        assert self.test_shape_3.processed_points[-1] == (100, 0, "straight")
        assert self.test_shape_3.processed_points[-2][2] == 'spline'
        assert self.test_shape_3.processed_points[-3][2] == 'spline'

    def test_convert_circles_to_splines_volume(self):
        """creates a RotateMixedShape with a circular edge and converts the
        edge to a spline edges. Checks the new shape has approximately the same
        volume as the original shape (with circles)"""

        original_volume = self.test_shape_2.volume()
        self.test_shape_2.convert_all_circle_connections_to_splines()
        new_volume = self.test_shape_2.volume()

        assert pytest.approx(new_volume, rel=0.000001) == original_volume

        original_volume = self.test_shape_3.volume()
        self.test_shape_3.convert_all_circle_connections_to_splines()
        new_volume = self.test_shape_3.volume()

        assert pytest.approx(new_volume, rel=0.00001) == original_volume


if __name__ == "__main__":
    unittest.main()




import unittest

import paramak


class TestVacuumVessel(unittest.TestCase):

    def setUp(self):
        self.test_shape = paramak.VacuumVessel(
            height=2, inner_radius=1, thickness=0.2
        )

    def test_creation(self):
        """Creates a shape using the VacuumVessel parametric component and
        checks that a cadquery solid is created."""

        assert self.test_shape.solid is not None

    def test_ports(self):
        """Creates a vacuum vessel with ports holes in it and checks that a
        caquery solid is created."""

        cutter1 = paramak.PortCutterRectangular(
            distance=3, center_point=(
                0, 0), height=0.2, width=0.4, fillet_radius=0.01)
        cutter2 = paramak.PortCutterRectangular(
            distance=3, center_point=(
                0.5, 0), height=0.2, width=0.4, fillet_radius=0.00)
        cutter3 = paramak.PortCutterRectangular(
            distance=3, center_point=(-0.5, 0), height=0.2, width=0.4,
            physical_groups=None)
        cutter4 = paramak.PortCutterCircular(
            distance=3,
            center_point=(
                0.25,
                0),
            radius=0.1,
            azimuth_placement_angle=45,
            physical_groups=None)
        cutter5 = paramak.PortCutterRotated(
            (0, 0), azimuth_placement_angle=-90, rotation_angle=10,
            fillet_radius=0.01, physical_groups=None)

        pre_cut_volume = self.test_shape.volume()

        self.test_shape.cut = [cutter1, cutter2, cutter3, cutter4, cutter5]
        assert self.test_shape.solid is not None
        assert self.test_shape.volume() < pre_cut_volume




import numpy as np
from paramak import RotateSplineShape


class Plasma(RotateSplineShape):
    """Creates a double null tokamak plasma shape that is controlled by 4
    shaping parameters.

    Args:
        elongation (float, optional): the elongation of the plasma.
            Defaults to 2.0.
        major_radius (float, optional): the major radius of the plasma (cm).
            Defaults to 450.0.
        minor_radius (int, optional): the minor radius of the plasma (cm).
            Defaults to 150.0.
        triangularity (float, optional): the triangularity of the plasma.
            Defaults to 0.55.
        vertical_displacement (float, optional): the vertical_displacement
            of the plasma (cm). Defaults to 0.0.
        num_points (int, optional): number of points to describe the
            shape. Defaults to 50.
        configuration (str, optional): plasma configuration
            ("non-null", "single-null", "double-null").
            Defaults to "non-null".
        x_point_shift (float, optional): shift parameters for locating the
            X points in [0, 1]. Defaults to 0.1.
        name (str, optional): Defaults to "plasma".
    """

    def __init__(
        self,
        elongation=2.0,
        major_radius=450.0,
        minor_radius=150.0,
        triangularity=0.55,
        vertical_displacement=0.0,
        num_points=50,
        configuration="non-null",
        x_point_shift=0.1,
        name="plasma",
        **kwargs
    ):

        super().__init__(
            name=name,
            **kwargs
        )

        # properties needed for plasma shapes
        self.elongation = elongation
        self.major_radius = major_radius
        self.minor_radius = minor_radius
        self.triangularity = triangularity
        self.vertical_displacement = vertical_displacement
        self.num_points = num_points
        self.configuration = configuration
        self.x_point_shift = x_point_shift

        self.outer_equatorial_point = None
        self.inner_equatorial_point = None
        self.high_point = None
        self.low_point = None

    @property
    def high_point(self):
        self.high_point = (
            self.major_radius - self.triangularity * self.minor_radius,
            self.elongation * self.minor_radius + self.vertical_displacement,
        )
        return self._high_point

    @high_point.setter
    def high_point(self, value):
        self._high_point = value

    @property
    def low_point(self):
        self.low_point = (
            self.major_radius - self.triangularity * self.minor_radius,
            -self.elongation * self.minor_radius + self.vertical_displacement,
        )
        return self._low_point

    @low_point.setter
    def low_point(self, value):
        self._low_point = value

    @property
    def outer_equatorial_point(self):
        self.outer_equatorial_point = (
            self.major_radius + self.minor_radius, self.vertical_displacement)
        return self._outer_equatorial_point

    @outer_equatorial_point.setter
    def outer_equatorial_point(self, value):
        self._outer_equatorial_point = value

    @property
    def inner_equatorial_point(self):
        self.inner_equatorial_point = (
            self.major_radius - self.minor_radius, self.vertical_displacement)
        return self._inner_equatorial_point

    @inner_equatorial_point.setter
    def inner_equatorial_point(self, value):
        self._inner_equatorial_point = value

    @property
    def lower_x_point(self):
        self.compute_x_points()
        return self._lower_x_point

    @lower_x_point.setter
    def lower_x_point(self, value):
        self._lower_x_point = value

    @property
    def upper_x_point(self):
        self.compute_x_points()
        return self._upper_x_point

    @upper_x_point.setter
    def upper_x_point(self, value):
        self._upper_x_point = value

    @property
    def vertical_displacement(self):
        return self._vertical_displacement

    @vertical_displacement.setter
    def vertical_displacement(self, value):
        self._vertical_displacement = value

    @property
    def minor_radius(self):
        return self._minor_radius

    @minor_radius.setter
    def minor_radius(self, value):
        if value < 1:
            raise ValueError("minor_radius is out of range")
        else:
            self._minor_radius = value

    @property
    def major_radius(self):
        return self._major_radius

    @major_radius.setter
    def major_radius(self, value):
        if value < 1:
            raise ValueError("major_radius is out of range")
        else:
            self._major_radius = value

    @property
    def elongation(self):
        return self._elongation

    @elongation.setter
    def elongation(self, value):
        if value > 10 or value < 0:
            raise ValueError("elongation is out of range")
        else:
            self._elongation = value

    def compute_x_points(self):
        """Computes the location of X points based on plasma parameters and
        configuration

        Returns:
            ((float, float), (float, float)): lower and upper x points
            coordinates. None if no x points
        """
        lower_x_point, upper_x_point = None, None  # non-null config
        shift = self.x_point_shift
        elongation = self.elongation
        triangularity = self.triangularity
        if self.configuration in ["single-null", "double-null"]:
            # no X points for non-null config
            lower_x_point = (1 -
                             (1 +
                              shift) *
                             triangularity *
                             self.minor_radius, -
                             (1 +
                              shift) *
                             elongation *
                             self.minor_radius +
                             self.vertical_displacement, )

            if self.configuration == "double-null":
                # upper_x_point is up-down symmetrical
                upper_x_point = (
                    lower_x_point[0],
                    (1 + shift) * elongation * self.minor_radius
                    + self.vertical_displacement,
                )
        self.lower_x_point = lower_x_point
        self.upper_x_point = upper_x_point

    def find_points(self):
        """Finds the XZ points that describe the 2D profile of the plasma."""

        # create array of angles theta
        theta = np.linspace(0, 2 * np.pi, num=self.num_points, endpoint=False)

        # parametric equations for plasma
        def R(theta):
            return self.major_radius + self.minor_radius * np.cos(
                theta + self.triangularity * np.sin(theta)
            )

        def Z(theta):
            return (
                self.elongation * self.minor_radius * np.sin(theta)
                + self.vertical_displacement
            )

        self.points = np.stack((R(theta), Z(theta)), axis=1).tolist()




from typing import Optional

from paramak import ExtrudeCircleShape


class PortCutterCircular(ExtrudeCircleShape):
    """Creates an extruded shape with a circular section that is used to cut
    other components (eg. blanket, vessel,..) in order to create ports.

    Args:
        radius: radius (cm) of port cutter.
        distance: extruded distance (cm) of the port cutter.
        center_point: center point of the port cutter. Defaults to (0, 0).
        workplane: workplane in which the port cutters are created. Defaults
            to "ZY".
        rotation_axis: axis around which the port cutters are rotated and
            placed. Defaults to "Z".
        extrusion_start_offset: the distance between 0 and the start of the
            extrusion. Defaults to 1..
        name: defaults to "circular_port_cutter".
    """

    def __init__(
        self,
        radius: float,
        distance: float,
        center_point: Optional[tuple] = (0, 0),
        workplane: Optional[str] = "ZY",
        rotation_axis: Optional[str] = "Z",
        extrusion_start_offset: Optional[float] = 1.,
        name: Optional[str] = "circular_port_cutter",
        **kwargs
    ):
        super().__init__(
            workplane=workplane,
            rotation_axis=rotation_axis,
            extrusion_start_offset=extrusion_start_offset,
            radius=radius,
            extrude_both=False,
            name=name,
            distance=distance,
            **kwargs
        )

        self.radius = radius
        self.center_point = center_point

    def find_points(self):
        self.points = [self.center_point]




import unittest

import paramak
import pytest


class TestSingleNullBallReactor(unittest.TestCase):
    """tests functionality of the test_SingleNullBallReactor class"""

    def setUp(self):
        self.test_reactor = paramak.SingleNullBallReactor(
            inner_bore_radial_thickness=50,
            inboard_tf_leg_radial_thickness=200,
            center_column_shield_radial_thickness=50,
            divertor_radial_thickness=100,
            inner_plasma_gap_radial_thickness=150,
            plasma_radial_thickness=100,
            outer_plasma_gap_radial_thickness=50,
            plasma_gap_vertical_thickness=50,
            firstwall_radial_thickness=50,
            blanket_radial_thickness=100,
            blanket_rear_wall_radial_thickness=10,
            elongation=2,
            triangularity=0.55,
            number_of_tf_coils=4,
            rear_blanket_to_tf_gap=10,
            outboard_tf_coil_radial_thickness=10,
            outboard_tf_coil_poloidal_thickness=10,
            divertor_position="lower",
            rotation_angle=180,
            pf_coil_case_thicknesses=[10, 10, 10, 10],
            pf_coil_radial_thicknesses=[20, 50, 50, 20],
            pf_coil_vertical_thicknesses=[20, 50, 50, 20],
            pf_coil_radial_position=[500, 575, 575, 500],
            pf_coil_vertical_position=[300, 100, -100, -300],
        )

    def test_input_variable_names(self):
        """tests that the number of inputs variables is correct"""

        assert len(self.test_reactor.input_variables.keys()) == 28
        assert len(self.test_reactor.input_variable_names) == 28

    def test_single_null_ball_reactor_with_pf_and_tf_coils(self):
        """Checks that a SingleNullBallReactor with optional pf and tf coils can
        be created and that the correct number of components are produced."""

        assert len(self.test_reactor.shapes_and_components) == 10

    def test_single_null_ball_reactor_rotation_angle_impacts_volume(self):
        """Creates SingleNullBallReactors with different rotation angles and
        checks that the relative volumes of the components are correct."""

        self.test_reactor.rotation_angle = 90
        test_reactor_90_components = self.test_reactor.shapes_and_components
        self.test_reactor.rotation_angle = 180
        test_reactor_180_components = self.test_reactor.shapes_and_components

        for r90, r180 in zip(test_reactor_90_components,
                             test_reactor_180_components):
            assert r90.volume() == pytest.approx(r180.volume() * 0.5, rel=0.1)

    def test_hash_value(self):
        """Creates a single null ball reactor and checks that all shapes in the
        reactor are created when .shapes_and_components is first called. Checks
        that when .shapes_and_components is called again with no changes to the
        reactor, the shapes in the reactor are not reconstructed and the
        previously constructed shapes are returned. Checks that when
        .shapes_and_components is called again with changes to the reactor, the
        shapes in the reactor are reconstructed and these new shapes are
        returned. Checks that the reactor_hash_value is only updated when the
        reactor is reconstruced."""

        assert self.test_reactor.reactor_hash_value is None
        for key in [
            "_plasma",
            "_inboard_tf_coils",
            "_center_column_shield",
            "_divertor",
            "_firstwall",
            "_blanket",
            "_blanket_rear_wall",
            "_pf_coil",
            "_pf_coils_casing",
            "_tf_coil"
        ]:
            assert key not in self.test_reactor.__dict__.keys()
        assert self.test_reactor.shapes_and_components is not None

        for key in [
            "_plasma",
            "_inboard_tf_coils",
            "_center_column_shield",
            "_divertor",
            "_firstwall",
            "_blanket",
            "_blanket_rear_wall",
            "_pf_coil",
            "_pf_coils_casing",
            "_tf_coil"
        ]:
            assert key in self.test_reactor.__dict__.keys()
        assert len(self.test_reactor.shapes_and_components) == 10
        assert self.test_reactor.reactor_hash_value is not None
        initial_hash_value = self.test_reactor.reactor_hash_value
        self.test_reactor.rotation_angle = 270
        assert self.test_reactor.reactor_hash_value == initial_hash_value
        assert self.test_reactor.shapes_and_components is not None
        assert self.test_reactor.reactor_hash_value != initial_hash_value




import os
import unittest
from pathlib import Path

import pytest
from paramak import RotateStraightShape, SweepMixedShape


class TestSweepMixedShape(unittest.TestCase):

    def setUp(self):
        self.test_shape = SweepMixedShape(
            points=[(-10, -10, "straight"), (-10, 10, "spline"), (0, 20, "spline"),
                    (10, 10, "circle"), (0, 0, "circle"), (10, -10, "straight")],
            path_points=[(50, 0), (30, 50), (70, 100), (50, 150)]
        )

    def test_default_parameters(self):
        """Checks that the default parameters of a SweepMixedShape are correct."""

        # assert self.test_shape.rotation_angle == 360
        assert self.test_shape.azimuth_placement_angle == 0
        assert self.test_shape.workplane == "XY"
        assert self.test_shape.path_workplane == "XZ"
        assert self.test_shape.force_cross_section == False

    def test_solid_construction_workplane(self):
        """Checks that SweepMixedShapes can be created in different workplanes"""

        self.test_shape.workplane = "YZ"
        self.test_shape.path_workplane = "YX"
        assert self.test_shape.solid is not None

        self.test_shape.workplane = "XZ"
        self.test_shape.path_workplane = "XY"
        assert self.test_shape.solid is not None

    def test_relative_shape_volume_points(self):
        """Creates two SweepMixedShapes and checks that their relative volumes
        are correct."""

        self.test_shape.points = [(-10, -10, "straight"), (-10, 10, "spline"), (0, 20, "spline"),
                                  (10, 10, "circle"), (0, 0, "circle"), (10, -10, "straight")]
        test_volume = self.test_shape.volume()
        self.test_shape.points = [(-20, -20, "straight"), (-20, 20, "spline"), (0, 40, "spline"),
                                  (20, 20, "circle"), (0, 0, "circle"), (20, -20, "straight")]
        assert self.test_shape.volume() == pytest.approx(
            test_volume * 4, rel=0.01)

    def test_relative_shape_volume_azimuthal_placement(self):
        """Creates two SweepMixedShapes and checks that their relative volumes
        are correct."""

        test_volume = self.test_shape.volume()
        self.test_shape.azimuth_placement_angle = [0, 90, 180, 270]
        assert self.test_shape.volume() == pytest.approx(
            test_volume * 4, rel=0.01)

    def test_workplane_path_workplane_error_raises(self):
        """Checks that errors are raised when SweepMixedShapes are created with
        disallowed workplane and path_workplane combinations."""

        def workplane_and_path_workplane_equal():
            self.test_shape.workplane = "XZ"
            self.test_shape.path_workplane = "XZ"

        def invalid_relative_workplane_and_path_workplane():
            self.test_shape.workplane = "XZ"
            self.test_shape.path_workplane = "YZ"

        self.assertRaises(ValueError, workplane_and_path_workplane_equal)
        self.assertRaises(
            ValueError,
            invalid_relative_workplane_and_path_workplane)

    def test_workplane_opposite_distance(self):
        """Checks that a SweepMixedShape can be created with workplane XZ and
        path_workplane XY."""

        self.test_shape.workplane = "XZ"
        self.test_shape.path_workplane = "XY"

        assert self.test_shape.solid is not None

    def test_force_cross_section(self):
        """Checks that a SweepMixedshape with the same cross-section at each path_point
        is created when force_cross_section = True."""

        self.test_shape.force_cross_section = True

        test_area = round(min(self.test_shape.areas))

        assert self.test_shape.areas.count(
            pytest.approx(test_area, rel=0.01)) == 2

        cutting_shape = RotateStraightShape(
            points=[(0, 50), (0, 200), (100, 200), (100, 50)]
        )
        self.test_shape.cut = cutting_shape

        assert self.test_shape.areas.count(
            pytest.approx(test_area, rel=0.01)) == 2

        cutting_shape.points = [(0, 100), (0, 200), (100, 200), (100, 100)]
        self.test_shape.cut = cutting_shape

        assert self.test_shape.areas.count(
            pytest.approx(test_area, rel=0.01)) == 2

    def test_force_cross_section_volume(self):
        """Checks that a SweepMixedShape with a larger volume is created when
        force_cross_section = True than when force_cross_section = False."""

        test_volume = self.test_shape.volume()
        self.test_shape.force_cross_section = True
        assert self.test_shape.volume() > test_volume

    def test_surface_count(self):
        """Creates a SweepStraightShape and checks that it has the correct number
        of surfaces."""

        assert len(self.test_shape.areas) == 6
        assert len(set(round(i) for i in self.test_shape.areas)) == 5

    def test_export_stp(self):
        """Exports and stp file with mode = solid and wire and checks
        that the outputs exist and relative file sizes are correct."""

        os.system("rm test_solid.stp test_solid2.stp test_wire.stp")

        self.test_shape.export_stp('test_solid.stp', mode='solid')
        self.test_shape.export_stp('test_solid2.stp')
        self.test_shape.export_stp('test_wire.stp', mode='wire')

        assert Path("test_solid.stp").exists() is True
        assert Path("test_solid2.stp").exists() is True
        assert Path("test_wire.stp").exists() is True

        assert Path("test_solid.stp").stat().st_size == \
            Path("test_solid2.stp").stat().st_size
        assert Path("test_wire.stp").stat().st_size < \
            Path("test_solid2.stp").stat().st_size

        os.system("rm test_solid.stp test_solid2.stp test_wire.stp")


if __name__ == "__main__":
    unittest.main()



import unittest

import paramak
import pytest


class TestToroidalFieldCoilCoatHanger(unittest.TestCase):
    def setUp(self):
        self.test_shape = paramak.ToroidalFieldCoilCoatHanger(
            horizontal_start_point=(200, 500),
            horizontal_length=400,
            vertical_mid_point=(700, 0),
            vertical_length=500,
            thickness=50,
            distance=30,
            number_of_coils=1,
            with_inner_leg=True,
        )

    def test_default_parameters(self):
        """Checks that the default parameters of a ToroidalFieldCoilCoatHanger
        are correct."""

        assert self.test_shape.rotation_angle == 360
        assert self.test_shape.with_inner_leg

    def test_points_calculation(self):
        """Checks that the points used to construct the
        ToroidalFieldCoilCoatHanger are calculated correctly from the parameters
        given."""

        assert self.test_shape.points == [
            (200, 500),
            (600, 500),
            (700, 250.0),
            (700, -250.0),
            (600, -500),
            (200, -500),
            (200, -550),
            (600, -550),
            (646.423834544263, -518.5695338177052),
            (746.423834544263, -268.5695338177052),
            (750, -250.0),
            (750, 250.0),
            (746.423834544263, 268.5695338177052),
            (646.423834544263, 518.5695338177052),
            (600, 550),
            (200, 550),
        ]

    def test_processed_points_calculation(self):
        """Checks that the processed_points used to construct the
        ToroidalFieldCoilCoatHanger are calculated correctly from the parameters
        given."""

        assert self.test_shape.processed_points == [
            (200, 500, "straight"),
            (600, 500, "straight"),
            (700, 250.0, "straight"),
            (700, -250.0, "straight"),
            (600, -500, "straight"),
            (200, -500, "straight"),
            (200, -550, "straight"),
            (600, -550, "straight"),
            (646.423834544263, -518.5695338177052, "straight"),
            (746.423834544263, -268.5695338177052, "straight"),
            (750, -250.0, "straight"),
            (750, 250.0, "straight"),
            (746.423834544263, 268.5695338177052, "straight"),
            (646.423834544263, 518.5695338177052, "straight"),
            (600, 550, "straight"),
            (200, 550, "straight"),
            (200, 500, "straight"),
        ]

    def test_creation_with_inner_leg(self):
        """Creates a tf coil with inner leg using the ToroidalFieldCoilCoatHanger
        parametric component and checks that a cadquery solid is created."""

        assert self.test_shape.solid is not None
        assert self.test_shape.volume() > 1000
        assert self.test_shape.inner_leg_connection_points is not None

        test_inner_leg = paramak.ExtrudeStraightShape(
            points=self.test_shape.inner_leg_connection_points, distance=30
        )
        assert test_inner_leg.solid is not None

    def test_creation_with_inner_leg_with_overlap(self):
        """Creates tf coils with overlapping inner legs using the ToroidalFieldCoilCoatHanger
        parametric component and checks that a cadquery solid is created correctly."""

        self.test_shape.horizontal_start_point = (0, 500)
        self.test_shape.number_of_coils = 8
        assert self.test_shape.solid is not None

        with_inner_leg_volume = self.test_shape.volume()

        self.test_shape.with_inner_leg = False
        without_inner_leg_volume = self.test_shape.volume()

        assert self.test_shape.solid is not None

        test_inner_leg = paramak.ExtrudeStraightShape(
            points=[(0, 500), (50, 500), (50, -500), (0, -500)],
            distance=30,
            azimuth_placement_angle=[0, 45, 90, 135, 180, 225, 270, 315],
        )
        inner_leg_volume = test_inner_leg.volume()

        assert with_inner_leg_volume == pytest.approx(
            without_inner_leg_volume + inner_leg_volume
        )

    def test_creation_no_inner_leg(self):
        """Creates a tf coil with no inner leg using the ToroidalFieldCoilRectangle
        parametric component and checks that a cadquery solid is created."""

        test_volume = self.test_shape.volume()

        test_inner_leg = paramak.ExtrudeStraightShape(
            points=self.test_shape.inner_leg_connection_points, distance=30
        )
        inner_leg_volume = test_inner_leg.volume()

        self.test_shape.with_inner_leg = False

        assert self.test_shape.solid is not None
        assert self.test_shape.volume() == pytest.approx(test_volume - inner_leg_volume)

    def test_absolute_volume(self):
        """Creates a tf coil using the ToroidalFieldCoilCoatHanger parametric
        component and checks that the volume is correc."""

        assert self.test_shape.volume() == pytest.approx(
            (400 * 50 * 30 * 2)
            + ((50 * 50 * 30 / 2) * 2)
            + (50 * 500 * 30)
            + (
                ((150 * 250 * 30) - (((100 * 250) / 2) * 30) - (((100 * 250) / 2) * 30))
                * 2
            )
            + (50 * 1000 * 30),
            rel=0.1,
        )

        self.test_shape.with_inner_leg = False
        assert self.test_shape.volume() == pytest.approx(
            (400 * 50 * 30 * 2)
            + ((50 * 50 * 30 / 2) * 2)
            + (50 * 500 * 30)
            + (
                ((150 * 250 * 30) - (((100 * 250) / 2) * 30) - (((100 * 250) / 2) * 30))
                * 2
            ),
            rel=0.1,
        )

        self.test_shape.with_inner_leg = True
        self.test_shape.number_of_coils = 8
        assert self.test_shape.volume() == pytest.approx(
            (
                (400 * 50 * 30 * 2)
                + ((50 * 50 * 30 / 2) * 2)
                + (50 * 500 * 30)
                + (
                    (
                        (150 * 250 * 30)
                        - (((100 * 250) / 2) * 30)
                        - (((100 * 250) / 2) * 30)
                    )
                    * 2
                )
                + (50 * 1000 * 30)
            )
            * 8,
            rel=0.1,
        )

        self.test_shape.with_inner_leg = False
        assert self.test_shape.volume() == pytest.approx(
            (
                (400 * 50 * 30 * 2)
                + ((50 * 50 * 30 / 2) * 2)
                + (50 * 500 * 30)
                + (
                    (
                        (150 * 250 * 30)
                        - (((100 * 250) / 2) * 30)
                        - (((100 * 250) / 2) * 30)
                    )
                    * 2
                )
            )
            * 8,
            rel=0.1,
        )

    def test_absolute_area(self):
        """Creates a tf coil using the ToroidalFieldCoilCoatHanger parametric
        component and checks that the areas of the faces are correct."""

        assert self.test_shape.area == pytest.approx(
            (
                (
                    (400 * 50 * 2)
                    + (50 * 50 * 0.5 * 2)
                    + (((150 * 250) - (100 * 250 * 0.5) - (100 * 250 * 0.5)) * 2)
                    + (500 * 50)
                )
                * 2
            )
            + ((50 * 30) * 4)
            + ((400 * 30) * 4)
            + ((500 * 30) * 2)
            + ((((50 ** 2 + 50 ** 2) ** 0.5) * 30) * 2)
            + ((((100 ** 2 + 250 ** 2) ** 0.5) * 30) * 4)
            + ((50 * 1000) * 2)
            + ((1000 * 30) * 2),
            rel=0.1,
        )
        assert len(self.test_shape.areas) == 24

        assert self.test_shape.areas.count(pytest.approx(50 * 30)) == 4
        assert self.test_shape.areas.count(pytest.approx(400 * 30)) == 4
        assert self.test_shape.areas.count(pytest.approx(500 * 30)) == 2
        assert (
            self.test_shape.areas.count(
                pytest.approx(((100 ** 2 + 250 ** 2) ** 0.5) * 30)
            )
            == 4
        )
        assert self.test_shape.areas.count(pytest.approx(50 * 1000)) == 2
        assert self.test_shape.areas.count(pytest.approx(1000 * 30)) == 2

        self.test_shape.with_inner_leg = False
        assert self.test_shape.area == pytest.approx(
            (
                (
                    (400 * 50 * 2)
                    + (50 * 50 * 0.5 * 2)
                    + (((150 * 250) - (100 * 250 * 0.5) - (100 * 250 * 0.5)) * 2)
                    + (500 * 50)
                )
                * 2
            )
            + ((50 * 30) * 2)
            + ((400 * 30) * 4)
            + ((500 * 30) * 2)
            + ((((50 ** 2 + 50 ** 2) ** 0.5) * 30) * 2)
            + ((((100 ** 2 + 250 ** 2) ** 0.5) * 30) * 4),
            rel=0.1,
        )

    def test_rotation_angle(self):
        """Creates a tf coil with a rotation_angle < 360 degrees and checks
        that the correct cut is performed and the volume is correct."""

        self.test_shape.number_of_coils = 8

        self.test_shape.rotation_angle = 360
        self.test_shape.workplane = "XZ"
        test_volume = self.test_shape.volume()
        self.test_shape.rotation_angle = 180
        assert self.test_shape.volume() == pytest.approx(test_volume * 0.5, rel=0.01)

        self.test_shape.rotation_angle = 360
        self.test_shape.workplane = "YZ"
        test_volume = self.test_shape.volume()
        self.test_shape.rotation_angle = 180
        assert self.test_shape.volume() == pytest.approx(test_volume * 0.5, rel=0.01)

        self.test_shape.rotation_angle = 360
        self.test_shape.workplane = "XY"
        self.test_shape.rotation_axis = "Y"
        test_volume = self.test_shape.volume()
        self.test_shape.rotation_angle = 180
        assert self.test_shape.volume() == pytest.approx(test_volume * 0.5)




from typing import Optional

import cadquery as cq
from paramak import RotateStraightShape


class VacuumVesselInnerLeg(RotateStraightShape):
    """A cylindrical vessel volume with constant thickness.

    Arguments:
        inner_height: height of the vessel.
        inner_radius: the inner radius of the vessel.
        inner_leg_radius: the inner radius of the inner leg.
        thickness: thickness of the vessel
    """

    def __init__(
        self,
        inner_height: float,
        inner_radius: float,
        inner_leg_radius: float,
        thickness: float,
        **kwargs
    ):
        self.inner_height = inner_height
        self.inner_radius = inner_radius
        self.inner_leg_radius = inner_leg_radius
        self.thickness = thickness
        super().__init__(
            **kwargs
        )

    @property
    def inner_height(self):
        return self._inner_height

    @inner_height.setter
    def inner_height(self, value):
        if not isinstance(value, (float, int)):
            raise ValueError(
                'VacuumVesselInnerLeg.inner_height must be a number. Not', value)
        if value <= 0:
            msg = (
                f'VacuumVesselInnerLeg.inner_height must be a positive number above 0. Not {value}')
            raise ValueError(msg)
        self._inner_height = value

    @property
    def inner_radius(self):
        return self._inner_radius

    @inner_radius.setter
    def inner_radius(self, value):
        if not isinstance(value, (float, int)):
            msg = (
                f'VacuumVesselInnerLeg.inner_radius must be a number. Not {value}')
            raise ValueError(msg)
        if value <= 0:
            msg = (
                f'VacuumVesselInnerLeg.inner_radius must be a positive number above 0. Not {value}')
            raise ValueError(msg)

        self._inner_radius = value

    def find_points(self):
        """Finds the XZ points joined by straight connections that describe the
            2D profile of the vessel shape."""
        thickness = self.thickness
        inner_radius = self.inner_radius
        inner_height = self.inner_height
        inner_leg_radius = self.inner_leg_radius

        point_1 = (inner_leg_radius + 2 * thickness +
                   inner_radius, thickness + (inner_height / 2.))
        point_2 = (0, thickness + (inner_height / 2.))
        point_3 = (0, -(thickness + (inner_height / 2.)))
        point_4 = (inner_leg_radius + 2 * thickness +
                   inner_radius, -(thickness + (inner_height / 2)))

        point_5 = (
            inner_leg_radius +
            thickness +
            inner_radius,
            inner_height /
            2)
        point_6 = (inner_leg_radius + thickness, inner_height / 2)
        point_7 = (inner_leg_radius + thickness, -inner_height / 2)
        point_8 = (inner_leg_radius + thickness +
                   inner_radius, -inner_height / 2)

        points_9 = (inner_leg_radius, thickness + inner_height / 2)
        points_10 = (inner_leg_radius, -(thickness + inner_height / 2))

        self.points = [
            point_1, point_2, point_3, point_4, point_5, point_6, point_7,
            point_8, points_9, points_10
        ]

    def create_solid(self):
        """Creates a 3d solid using points with straight edges. Individual
        solids in the compound can be accessed using .Solids()[i] where i is an
        int

           Returns:
              A CadQuery solid: A 3D solid volume
        """

        local_points = []
        for point in self.points:
            local_points.append((point[0], point[1]))

        big_wire = (
            cq.Workplane(self.workplane)
            .polyline(local_points[:4])
        ).close()

        small_wire = (
            cq.Workplane(self.workplane)
            .polyline(local_points[4:8])  # list of points has 10 entries
        ).close()

        inner_wire = (
            cq.Workplane(self.workplane)
            .polyline(
                [
                    local_points[1],
                    local_points[2],
                    local_points[9],
                    local_points[8]
                ]
            )
        ).close()

        inner_solid = inner_wire.revolve(self.rotation_angle)
        big_solid = big_wire.revolve(self.rotation_angle)
        small_solid = small_wire.revolve(self.rotation_angle)

        solid = big_solid.cut(small_solid).cut(inner_solid)

        self.wire = [big_wire, small_wire]
        solid = self.rotate_solid(solid)
        solid = self.perform_boolean_operations(solid)

        self.solid = solid

        return solid




from typing import Optional, Tuple

from cadquery import Workplane
from paramak import Shape
from paramak.utils import calculate_wedge_cut


class ExtrudeCircleShape(Shape):
    """Extrudes a circular 3d CadQuery solid from a central point and a radius

    Args:
        distance: the extrusion distance to use (cm units if used for
            neutronics)
        radius: radius of the shape.
        rotation_angle: rotation_angle of solid created. a cut is performed
            from rotation_angle to 360 degrees. Defaults to 360.
        extrude_both: if set to True, the extrusion will occur in both
            directions. Defaults to True.
    """

    def __init__(
        self,
        distance: float,
        radius: float,
        extrusion_start_offset: Optional[float] = 0.0,
        rotation_angle: Optional[float] = 360,
        extrude_both: Optional[bool] = True,
        color: Optional[Tuple[float, float, float, Optional[float]]] = (0.984, 0.603, 0.6),
        **kwargs
    ):

        super().__init__(
            color=color,
            **kwargs
        )

        self.radius = radius
        self.distance = distance
        self.rotation_angle = rotation_angle
        self.extrude_both = extrude_both
        self.extrusion_start_offset = extrusion_start_offset

    @property
    def radius(self):
        return self._radius

    @radius.setter
    def radius(self, value):
        self._radius = value

    @property
    def distance(self):
        return self._distance

    @distance.setter
    def distance(self, value):
        self._distance = value

    @property
    def rotation_angle(self):
        return self._rotation_angle

    @rotation_angle.setter
    def rotation_angle(self, value):
        self._rotation_angle = value

    @property
    def extrusion_start_offset(self):
        return self._extrusion_start_offset

    @extrusion_start_offset.setter
    def extrusion_start_offset(self, value):
        self._extrusion_start_offset = value

    def create_solid(self):
        """Creates a extruded 3d solid using points with circular edges.

           Returns:
              A CadQuery solid: A 3D solid volume
        """

        # so a positive offset moves extrusion further from axis of azimuthal
        # placement rotation
        extrusion_offset = -self.extrusion_start_offset

        if not self.extrude_both:
            extrusion_distance = -self.distance
        else:
            extrusion_distance = -self.distance / 2.0

        wire = (
            Workplane(self.workplane)
            .workplane(offset=extrusion_offset)
            .moveTo(self.points[0][0], self.points[0][1])
            .circle(self.radius)
        )

        self.wire = wire

        solid = wire.extrude(
            distance=extrusion_distance,
            both=self.extrude_both)

        solid = self.rotate_solid(solid)
        cutting_wedge = calculate_wedge_cut(self)
        solid = self.perform_boolean_operations(solid, wedge_cut=cutting_wedge)
        self.solid = solid

        return solid




import math
import unittest

import paramak
import pytest


class TestPoloidalFieldCoilCaseFC(unittest.TestCase):

    def setUp(self):
        self.pf_coil = paramak.PoloidalFieldCoil(
            height=50, width=60, center_point=(1000, 500)
        )

        self.test_shape = paramak.PoloidalFieldCoilCaseFC(
            pf_coil=self.pf_coil, casing_thickness=5
        )

    def test_default_parameters(self):
        """Checks that the default parameters of a PoloidalFieldCoilCaseFC are correct."""

        assert self.test_shape.rotation_angle == 360

    def test_creation(self):
        """Creates a pf coil case using the PoloidalFieldCoilCaseFC parametric
        component and checks that a cadquery solid is created."""

        assert self.test_shape.solid is not None
        assert self.test_shape.volume() > 1000

    def test_absolute_volume(self):
        """Creates a pf coil case using the PoloidalFieldCoilCaseFC parametric
        component and checks that its volume is correct."""

        assert self.test_shape.volume() == pytest.approx(
            (math.pi * 2 * 1000) * ((50 * 5 * 2) + (70 * 5 * 2)))

    def test_absolute_areas(self):
        """Creates a pf coil case using the PoloidalFieldCoilCaseFC parametric
        component and checks that the areas of its faces are correct"""

        assert len(self.test_shape.areas) == 8
        assert len(set([round(i) for i in self.test_shape.areas])) == 6
        assert self.test_shape.areas.count(
            pytest.approx(60 * math.pi * 2 * 1000)) == 2
        assert self.test_shape.areas.count(
            pytest.approx(70 * math.pi * 2 * 1000)) == 2
        assert self.test_shape.areas.count(
            pytest.approx(50 * math.pi * 2 * 1030)) == 1
        assert self.test_shape.areas.count(
            pytest.approx(50 * math.pi * 2 * 970)) == 1
        assert self.test_shape.areas.count(
            pytest.approx(60 * math.pi * 2 * 1035)) == 1
        assert self.test_shape.areas.count(
            pytest.approx(60 * math.pi * 2 * 965)) == 1




import unittest

import paramak


class TestBlanketFP(unittest.TestCase):
    def test_creation(self):
        blanket = paramak.BlanketFPPoloidalSegments(
            thickness=20, start_angle=0,
            stop_angle=180, rotation_angle=180)
        assert blanket.solid is not None

    def test_creation_with_optimiser(self):
        # Default
        blanket = paramak.BlanketFPPoloidalSegments(
            thickness=20, start_angle=0,
            stop_angle=180, rotation_angle=180)
        assert blanket.solid is not None

        # With limits
        blanket.length_limits = (10, 300)
        blanket.nb_segments_limits = (4, 8)
        assert blanket.solid is not None

        # With None length_limits
        blanket.length_limits = None
        blanket.nb_segments_limits = (4, 8)
        assert blanket.solid is not None

        # With None nb_segments_limits
        blanket.start_angle = 80
        blanket.length_limits = (100, 300)
        blanket.nb_segments_limits = None
        assert blanket.solid is not None

    def test_optimiser(self):
        blanket = paramak.BlanketFPPoloidalSegments(
            thickness=20, start_angle=0,
            stop_angle=180, rotation_angle=180)

        blanket.length_limits = (100, 300)
        blanket.nb_segments_limits = (2, 8)
        assert blanket.solid is not None

        def no_possible_config():
            blanket.length_limits = (10, 20)
            blanket.nb_segments_limits = (2, 4)
            blanket.solid

        self.assertRaises(ValueError, no_possible_config)

    def test_modifying_nb_segments_limits(self):
        """creates a shape and checks that modifying the nb_segments_limits
        also modifies segmets_angles accordingly
        """
        blanket1 = paramak.BlanketFPPoloidalSegments(
            thickness=20, start_angle=0,
            stop_angle=180, rotation_angle=180)

        blanket2 = paramak.BlanketFPPoloidalSegments(
            thickness=20, start_angle=0,
            stop_angle=180, rotation_angle=180,
            cut=blanket1)

        blanket1.nb_segments_limits = (4, 8)
        blanket2.nb_segments_limits = (3, 8)

        assert blanket2.volume() != 0

    def test_segments_angles_is_modified_num_segments(self):
        blanket1 = paramak.BlanketFPPoloidalSegments(
            thickness=20, start_angle=0,
            stop_angle=180, rotation_angle=180)

        blanket2 = paramak.BlanketFPPoloidalSegments(
            thickness=20, start_angle=0,
            stop_angle=180, rotation_angle=180,
            cut=blanket1)

        blanket1.num_segments = 8
        blanket2.num_segments = 5
        assert blanket2.volume() != 0

    def test_num_point_is_affected(self):
        blanket = paramak.BlanketFPPoloidalSegments(
            thickness=20, start_angle=0,
            stop_angle=180, rotation_angle=180)
        assert blanket.num_points == blanket.num_segments + 1
        blanket.num_segments = 60
        assert blanket.num_points == blanket.num_segments + 1

    def test_segment_angles_affects_solid(self):
        blanket1 = paramak.BlanketFPPoloidalSegments(
            thickness=20, start_angle=0,
            stop_angle=180, rotation_angle=180)
        blanket2 = paramak.BlanketFPPoloidalSegments(
            thickness=20, start_angle=0,
            stop_angle=180, rotation_angle=180,
            cut=blanket1)
        blanket2.segments_angles = [0, 25, 50, 90, 130, 150, 180]
        assert blanket2.volume() != 0

    def test_warning_segment_angles(self):
        blanket = paramak.BlanketFPPoloidalSegments(
            thickness=20, start_angle=1,
            stop_angle=50, rotation_angle=180)

        def warning1():
            blanket.start_angle = 1
            blanket.stop_angle = 50
            blanket.num_segments = None
            blanket.segments_angles = [0, 25, 50, 90, 130, 150, 180]

        def warning2():
            blanket.start_angle = None
            blanket.stop_angle = None
            blanket.num_segments = 7
            blanket.segments_angles = [0, 25, 50, 90, 130, 150, 180]

        self.assertWarns(UserWarning, warning1)
        self.assertWarns(UserWarning, warning2)

    def test_creation_with_gaps(self):
        blanket = paramak.BlanketFPPoloidalSegments(
            thickness=20, start_angle=0,
            stop_angle=180, rotation_angle=180,
            segments_gap=3
        )
        assert blanket.solid is not None




from typing import Optional, Tuple

from paramak import RotateStraightShape


class PoloidalFieldCoilCaseFC(RotateStraightShape):
    """Creates a casing for a rectangular poloidal field coil by building
    around an existing coil (which is passed as an argument on construction).

    Args:
        pf_coil (paramak.PoloidalFieldCoil): a pf coil object with a set width,
            height and center point.
        casing_thickness (float): the thickness of the coil casing (cm).
    """

    def __init__(
        self,
        pf_coil,
        casing_thickness,
        color: Optional[Tuple[int, int, int]] = (1., 1., 0.498),
        **kwargs
    ):

        super().__init__(
            color=color,
            **kwargs
        )

        self.center_point = pf_coil.center_point
        self.height = pf_coil.height
        self.width = pf_coil.width
        self.casing_thickness = casing_thickness

    @property
    def center_point(self):
        return self._center_point

    @center_point.setter
    def center_point(self, value):
        self._center_point = value

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, height):
        self._height = height

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, width):
        self._width = width

    def find_points(self):
        """Finds the XZ points joined by straight connections that describe
        the 2D profile of the poloidal field coil case shape."""

        points = [
            (
                self.center_point[0] + self.width / 2.0,
                self.center_point[1] + self.height / 2.0,
            ),  # upper right
            (
                self.center_point[0] + self.width / 2.0,
                self.center_point[1] - self.height / 2.0,
            ),  # lower right
            (
                self.center_point[0] - self.width / 2.0,
                self.center_point[1] - self.height / 2.0,
            ),  # lower left
            (
                self.center_point[0] - self.width / 2.0,
                self.center_point[1] + self.height / 2.0,
            ),  # upper left
            (
                self.center_point[0] + self.width / 2.0,
                self.center_point[1] + self.height / 2.0,
            ),  # upper right
            (
                self.center_point[0] + \
                (self.casing_thickness + self.width / 2.0),
                self.center_point[1] + \
                (self.casing_thickness + self.height / 2.0),
            ),
            (
                self.center_point[0] + \
                (self.casing_thickness + self.width / 2.0),
                self.center_point[1] - \
                (self.casing_thickness + self.height / 2.0),
            ),
            (
                self.center_point[0] - \
                (self.casing_thickness + self.width / 2.0),
                self.center_point[1] - \
                (self.casing_thickness + self.height / 2.0),
            ),
            (
                self.center_point[0] - \
                (self.casing_thickness + self.width / 2.0),
                self.center_point[1] + \
                (self.casing_thickness + self.height / 2.0),
            ),
            (
                self.center_point[0] + \
                (self.casing_thickness + self.width / 2.0),
                self.center_point[1] + \
                (self.casing_thickness + self.height / 2.0),
            )
        ]

        self.points = points




import unittest
import pytest
import paramak


class TestVacuumVessel(unittest.TestCase):

    def setUp(self):
        self.cutter = paramak.PortCutterRectangular(
            width=100,
            height=100,
            distance=1100,
            azimuth_placement_angle=[
                0,
                90,
                180,
                270])
        self.test_shape = paramak.VacuumVesselInnerLeg(
            inner_height=1100,
            inner_radius=700,
            inner_leg_radius=100,
            thickness=100,
        )
        self.test_shape_cut = paramak.VacuumVesselInnerLeg(
            inner_height=1100,
            inner_radius=700,
            inner_leg_radius=100,
            thickness=100,
            cut=self.cutter
        )

    def test_creation(self):
        """Creates a shape using the VacuumVessel parametric component and
        checks that a cadquery solid is created."""

        assert self.test_shape.solid is not None

    def test_cut(self):
        assert self.test_shape.volume() != self.test_shape_cut.volume()




from typing import Optional
from paramak import Plasma, RotateMixedShape


class CenterColumnShieldPlasmaHyperbola(RotateMixedShape):
    """A center column shield volume with a curvature controlled by the shape
    of the plasma and offsets specified at the plasma center and edges. Shield
    thickness is controlled by the relative values of the shield offsets and
    inner radius.

    Args:
        height: height of the center column shield.
        inner_radius: the inner radius of the center column shield.
        mid_offset: the offset of the shield from the plasma at the plasma
            center.
        edge_offset: the offset of the shield from the plasma at the plasma
            edge.
        major_radius: the major radius of the plasma. Defaults to 450.0.
        minor_radius: the minor radius of the plasma. Defaults to 150.0.
        triangularity: the triangularity of the plasma. Defaults to 0.55.
        elongation: the elongation of the plasma. Defaults to 2.0.
    """

    def __init__(
        self,
        height: float,
        inner_radius: float,
        mid_offset: float,
        edge_offset: float,
        major_radius: Optional[float] = 450.0,
        minor_radius: Optional[float] = 150.0,
        triangularity: Optional[float] = 0.55,
        elongation: Optional[float] = 2.0,
        **kwargs
    ) -> None:

        super().__init__(**kwargs)

        self.major_radius = major_radius
        self.minor_radius = minor_radius
        self.triangularity = triangularity
        self.elongation = elongation
        self.height = height
        self.inner_radius = inner_radius
        self.mid_offset = mid_offset
        self.edge_offset = edge_offset

    @property
    def major_radius(self):
        return self._major_radius

    @major_radius.setter
    def major_radius(self, value):
        self._major_radius = value

    @property
    def minor_radius(self):
        return self._minor_radius

    @minor_radius.setter
    def minor_radius(self, value):
        self._minor_radius = value

    @property
    def triangularity(self):
        return self._triangularity

    @triangularity.setter
    def triangularity(self, value):
        self._triangularity = value

    @property
    def elongation(self):
        return self._elongation

    @elongation.setter
    def elongation(self, value):
        self._elongation = value

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, value):
        self._height = value

    @property
    def inner_radius(self):
        return self._inner_radius

    @inner_radius.setter
    def inner_radius(self, value):
        self._inner_radius = value

    @property
    def mid_offset(self):
        return self._mid_offset

    @mid_offset.setter
    def mid_offset(self, value):
        self._mid_offset = value

    @property
    def edge_offset(self):
        return self._edge_offset

    @edge_offset.setter
    def edge_offset(self, value):
        self._edge_offset = value

    def find_points(self):
        """Finds the XZ points and connection types (straight and spline) that
        describe the 2D profile of the center column shield shape."""

        plasma = Plasma()

        plasma.major_radius = self.major_radius
        plasma.minor_radius = self.minor_radius
        plasma.triangularity = self.triangularity
        plasma.elongation = self.elongation
        plasma.rotation_angle = self.rotation_angle
        plasma.find_points()

        plasma_height = abs(plasma.high_point[1]) + abs(plasma.low_point[1])
        if self.height <= plasma_height:
            msg = (f'Center column height ({self.height}) is smaller than '
                   'plasma height ({plasma_height}')
            raise ValueError(msg)

        if self.inner_radius >= plasma.inner_equatorial_point[0] - \
                self.mid_offset:
            raise ValueError("Inner radius is too large")

        points = [
            (self.inner_radius, 0, "straight"),
            (self.inner_radius, self.height / 2, "straight"),
            (
                plasma.high_point[0] - self.edge_offset,
                self.height / 2, "straight"
            ),
            (
                plasma.high_point[0] - self.edge_offset,
                plasma.high_point[1], "spline"
            ),
            (
                plasma.inner_equatorial_point[0] - self.mid_offset,
                plasma.inner_equatorial_point[1],
                "spline",
            ),
            (
                plasma.low_point[0] - self.edge_offset,
                plasma.low_point[1], "straight"),
            (
                plasma.low_point[0] - self.edge_offset,
                -1 * self.height / 2, "straight"
            ),
            (self.inner_radius, -1 * self.height / 2, "straight")
        ]

        self.points = points



from typing import Optional, Tuple
from paramak import RotateStraightShape


class CenterColumnShieldCylinder(RotateStraightShape):
    """A cylindrical center column shield volume with constant thickness.

    Args:
        height: height of the center column shield.
        inner_radius: the inner radius of the center column shield.
        outer_radius: the outer radius of the center column shield.
        center_height: the vertical height of the center of the component.
    """

    def __init__(
        self,
        height: float,
        inner_radius: float,
        outer_radius: float,
        center_height: Optional[float] = 0,
        name: Optional[str] = "CenterColumnShieldCylinder",
        color: Optional[Tuple[float, float, float, Optional[float]]] = (
            0.0,
            0.333,
            0.0,
        ),
        **kwargs
    ) -> None:

        super().__init__(
            color=color,
            name=name,
            **kwargs
        )

        self.height = height
        self.inner_radius = inner_radius
        self.outer_radius = outer_radius
        self.center_height = center_height

    @property
    def center_height(self):
        return self._center_height

    @center_height.setter
    def center_height(self, value):
        if not isinstance(value, (int, float)):
            msg = (
                f'CenterColumnShieldBlock.center_height should be a float or int. Not a {type(value)}')
            raise TypeError(msg)

        self._center_height = value

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, value):
        if value is None:
            raise ValueError(
                "height of the CenterColumnShieldBlock cannot be None")
        self._height = value

    @property
    def inner_radius(self):
        return self._inner_radius

    @inner_radius.setter
    def inner_radius(self, value):
        if hasattr(self, "outer_radius"):
            if value >= self.outer_radius:
                msg = (f'inner_radius ({value}) is larger than outer_radius '
                       '({self.outer_radius})')
                raise ValueError(msg)
        self._inner_radius = value

    @property
    def outer_radius(self):
        return self._outer_radius

    @outer_radius.setter
    def outer_radius(self, value):
        if hasattr(self, "inner_radius"):
            if value <= self.inner_radius:
                msg = (f'inner_radius ({self.inner_radius}) is larger than '
                       'outer_radius ({value})')
                raise ValueError(msg)

        self._outer_radius = value

    def find_points(self):
        """Finds the XZ points joined by straight connections that describe the
            2D profile of the center column shield shape."""

        points = [
            (self.inner_radius, self.center_height + self.height / 2),
            (self.outer_radius, self.center_height + self.height / 2),
            (self.outer_radius, self.center_height + (-self.height / 2)),
            (self.inner_radius, self.center_height + (-self.height / 2)),
        ]

        self.points = points
